from bluebeam.objects import PhysObject
import pygame


class Bullet(PhysObject):
    def __init__(self, g_, l_, pos_):
        super().__init__(g_, l_, pos_)
        self._objid = 3
        self.speed = pygame.math.Vector2(500, 500)
        self.vel = self.speed
        self.size = 32
        self.owner = 0  # Constant value (unless we add reflecting enemy/player shots.) 0 = player bullet, 1 = enemy bullet.
        self.knf = 0  # Kill Next Frame
        self._lifespan = 3.
        self._dmg = 10
        self._collide_terrain = True
        pass

    def update(self):
        if self.knf == 1:
            self.hit()
        else:
            self.update_phys()
            self.lifespan -= self.g.delta_time
            if self.lifespan <= 0:
                self.hit()
            if ((self.l.view + self.l.view_wh) - self.pos).length() > self.l.view_wh.length() * 1.75:
                self.hit()

    def render(self):
        pass
    
    @property
    def collide_terrain(self):
        return self._collide_terrain

    @collide_terrain.setter
    def collide_terrain(self, col):
        self._collide_terrain = col

    @property
    def lifespan(self):
        return self._lifespan

    @lifespan.setter
    def lifespan(self, lifespan_):
        self._lifespan = lifespan_

    @property
    def dmg(self):
        return self._dmg

    @dmg.setter
    def dmg(self, dmg_):
        self._dmg = dmg_

    @property
    def speed(self):
        return self._speed

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, size_):
        self._size = size_
        self.rect = pygame.rect.Rect(0, 0, self._size, self._size)
        self.rect.center = self.pos

    @speed.setter
    def speed(self, speed_):
        self._speed = speed_
        self._vel = self._speed

    def collision_target(self):
        if self.owner == 0:
            col = pygame.sprite.spritecollide(self, self.l.enemies, False)
        else:
            col = pygame.sprite.collide_rect(self, self.l.player)
        if col:
            return 1
        else:
            return 0

    def collision_surface(self):
        col = pygame.sprite.spritecollide(self, self.l.terrain, False)
        if len(col) > 0:
            return 1
        else:
            return 0

    def collision_terrain(self, vertical=None):
        if self.collision_target() or self.collision_surface():
            if(self._collide_terrain): self.knf = 1
            return 0
        else:
            return 0

    def hit(self):
        self.l.uncreate(self.instid)



import pygame
from bluebeam.objects import EnemyBullet

#boy, you can just SMELL the iji final boss rip-off-ing from miles away

class GammaBGRocket(EnemyBullet):

    def __init__(self,g_, l_, pos_):
        super().__init__(g_, l_, pos_)
        print("BGrocket spawned")
        self.spawnpos = pygame.math.Vector2()
        self.fakepos = pygame.math.Vector2()
        self.plr = self.l.player
        self.rect = pygame.rect.Rect(self.pos.x,self.pos.y,0,0)
        self.distance = 2.2
        self.depth = 16

    def setup(self, pos_):
        self.spawnpos = self.fakepos = pos_

    def update(self):
        self.distance -= self.g.delta_time
        self.fakepos = pygame.math.Vector2.lerp(self.spawnpos, self.pos, pow(self.clamp01((1.2-self.distance) / 1.2),3))
        if self.distance <= 0:
            self.boom()

    def clamp01(self, _c):
        if _c > 1: return 1
        elif _c < 0: return 0
        else: return _c

    def boom(self):
        boom = self.l.create("Explosion",self.pos.xy)
        boom.radius(256)
        boom.depth = 12
        self.l.uncreate(self.instid)
        pass

    def render(self):
        if self.distance < 1.2:
            p = pow(self.clamp01((1.2-self.distance) / 1.2),4)
            pygame.draw.circle(surface=self.l.s, color=(160*pow(p,0.33), 0, 0, 255), center=(self.fakepos - self.l.view), radius=32 + 64*p)
        pygame.draw.circle(surface=self.l.s, color=(255, 0, 0, 255), center=(self.pos - self.l.view), radius=128, width=24)


from bluebeam.objects import GameObject
import pygame
import math


# PHYSICS OBJECTS
# Physics objects are game objects with physics.
# Physics objects have the following attributes:
## Velocity (Vec2, objects should move by this much per tick as long as they aren't blocked)
## Gravity (float, added to velocity per tick when in air and not hovering)
## Physics State (int)
### 0: Ground State (Object is on ground)
### 1: Air State (Object is not on ground)
### 2: Hover State (Object ignores gravity entirely)

# I'm going to need assistance with handling collisions - specifically, actually detecting them. 
# Once I've figured out how to do that or someone else has gotten to it, then I can implement the physics update function.
# If anyone else wants to do that, make sure to check for horizontal collisions before vertical ones.

class PhysObject(GameObject):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self._objid = 1
        self._vel = pygame.math.Vector2(0, 0)
        self._gravity = 0
        self._pstate = 0
        self._speed = pygame.math.Vector2(200,200)
        self._collide_terrain = True

    @property
    def vel(self):
        return self._vel

    @vel.setter
    def vel(self, vel_):
        if isinstance(vel_, pygame.math.Vector2): self._vel = vel_

    @property
    def speed(self):
        return self._speed

    @speed.setter
    def speed(self, speed_):
        self._speed = speed_

    @property
    def gravity(self):
        return self._gravity

    @gravity.setter
    def gravity(self, gravity_):
        self._gravity = gravity_

    @property
    def pstate(self):
        return self._pstate

    @pstate.setter
    def pstate(self, pstate_):
        self._pstate = pstate_

    @property
    def collide_terrain(self):
        return self._collide_terrain

    @collide_terrain.setter
    def collide_terrain(self, col):
        self._collide_terrain = col

    def update_phys(self, gravity_mod=1):
        if (self.pstate == 1):
            self.vel.y += self.gravity * gravity_mod
        ovel = pygame.math.Vector2(self.vel.x, self.vel.y)
        self.vel *= self.g.delta_time

        # Physics calculations go here.

        # Collision checks: start with horizontal velocity
        xv_bounce = self.ccheck_h()
        self.pos.x += self.vel.x - math.modf(self.vel.x)[0]
        self.rect.center = self.pos

        # Next, vertical velocity
        yv_reset = self.ccheck_v()
        self.pos.y += self.vel.y - math.modf(self.vel.y)[0]
        self.rect.center = self.pos

        self.vel = ovel
        if (yv_reset): self.vel.y = 0
        if (xv_bounce): self.vel.x = 0

    def update(self):
        self.update_phys()

    # Horizontal movement collision checking
    def ccheck_h(self):
        if(self.collide_terrain==False): return False
        
        xv_bounce = False
        cc = self.rect.copy()
        for i in range(1, math.ceil(abs(self.vel.x))):
            self.rect = self.rect.move(math.copysign(1, self.vel.x), 0)
            if self.collision_terrain():
                xv_bounce = True
                self.vel = pygame.math.Vector2(math.copysign(i - 1, self.vel.x), self.vel.y)
                break
        self.rect = cc
        return xv_bounce

    # Vertical movement collision checking
    def ccheck_v(self):
        if(self.collide_terrain==False): return False

        yv_reset = False
        if (self.pstate == 0): self.pstate = 1
        cc = self.rect.copy()
        for i in range(1, math.ceil(abs(self.vel.y))):
            self.rect = self.rect.move(0, math.copysign(1, self.vel.y))
            if self.collision_terrain(True):
                yv_reset = True
                self.vel = pygame.math.Vector2(self.vel.x, math.copysign(i - 1, self.vel.y))
                break
        self.rect = cc
        return yv_reset

    def collision_terrain(self, vertical=False):
        col = pygame.sprite.spritecollide(self, self.l.terrain, False)
        if len(col) > 0:
            # Update pstate to determine if the PhysObject is grounded
            if (vertical): self.check_grounded(col)
            return 1
        else:
            return 0

    def check_grounded(self, col):
        if (self.pstate == 1):
            for c in col:
                if ( self.rect.bottom - 1 <= c.rect.top ):
                    # Player is grounded
                    self.pstate = 0
                    
                elif( self.rect.top < c.rect.top ):
                    # If we are here, we are partially stuck in the ground
                    # As such, we'll correctively increase our y position by 1 until we aren't stuck
                    self.pos.y -= 1



import math
from bluebeam.objects import Enemy, Laser
import time
import pygame, os
from pygame import Vector2 as vec2
from bluebeam.GlobalVars import GlobalVars

# BOSS 1: Alpha "Whale
# phase 1: 6 turrets + 1 core
## 4 flex turrets: shoot directly at the player
## 2 side turrets: extend downward and fire close to the ground
# phase 2: Y turrets + 2 cores
## 2 flex turrets
## 1 side turret
## 2 cannon turrets: shoot explosive projectiles at the character
## 1 laser turret: travels horizontally and periodically fires a beam
# phase 3: Z turrets + 2 cores + 2 metal shields + commander
## 2 metal shields: destructible parts that protect cores from below
## commander is functionally an invincible turret with more complex attacks

spawns = [[], [], []]

# phase 1 spawning targets
spawns[0] = [["WhaleCore", vec2(0, 0)],
             ["WhaleTurretFlex", vec2(-320, -50), 0],
             ["WhaleTurretFlex", vec2(-80, -100), 1.33],
             ["WhaleTurretFlex", vec2(80, -100), 1.33],
             ["WhaleTurretFlex", vec2(320, -50), 0],
             ["WhaleTurretSide", vec2(-400, -50), 2.165, 0],
             ["WhaleTurretSide", vec2(400, -50), 0, 180]]

# phase 2 spawning targets
spawns[1] = [["WhaleCore", vec2(-320, 32)],
             ["WhaleCore", vec2(320, -32)],
             ["WhaleTurretRocket", vec2(-225, -75), 0],
             ["WhaleTurretFlex", vec2(-75, 0), 0],
             ["WhaleTurretFlex", vec2(75, 0), 1.33],
             ["WhaleTurretRocket", vec2(225, -75), 2.17],
             ["WhaleTurretLaser", vec2(0, 75), 1.33]]

# phase 3 spawning targets
spawns[2] = [["WhaleCore", vec2(-320, 0)],
             ["WhaleCore", vec2(-200, -100)],
             ["WhaleTurretFlex", vec2(100, -100), 1.33],
             ["WhaleTurretFlex", vec2(50, 100), 0],
             ["WhaleCommander", vec2(320, -48)], ]


class AlphaWhale(Enemy):
    phase = 4  # phases count in reverse, starts at 4 for the first frame
    phase_tick = 0  # tickdown for phase changes, for the moving animation...
    cores = pygame.sprite.Group()  # contains only "WhaleCore enemies
    turrets = pygame.sprite.Group()

    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self._objid = 2
        self._hp = 1
        self._mhp = 1
        self._contactdmg = 0
        self.tpos = pos_
        self.floattick = 0
        self.rect = pygame.rect.Rect(self.pos.x, self.pos.y, 0, 0)
        self.rect.center = self.pos
        self.frect = pygame.rect.Rect(self.pos.x, self.pos.y, 960, 360)
        self.frect.center = self.pos
        self.floatoffset = -4000
        self.deathstate = 0
        self.deathoffset = pygame.math.Vector2()
        self.fadetimer = 0

        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images/boss/AlphaWhale")
        self.spritesheet = pygame.image.load(os.path.join(self.imageDir, 'boss01_body.png')).convert()
        self.load_sprites()

        pass

    def load_sprites(self):
        img = self.load_single((0,0,320,48))
        self.image = pygame.transform.scale(img, (3200,480))

    def update(self):
        if(self._active):
            self.pos = vec2(self.floatoffset,0) + self.deathoffset + self.tpos + vec2(18*math.cos(self.floattick * 4.5), 8*math.sin(self.floattick*2.25))
            self.rect.center = self.frect.center = self.pos
            self.floattick += self.g.delta_time
            self._hp = 0
            if abs(self.floatoffset) > 0:
                sub = self.g.delta_time * 600
                self.floatoffset = math.copysign(max(abs(self.floatoffset) - sub, 0),self.floatoffset)
            for core in self.cores:
                self._hp += core.hp
            if self._hp <= 0 and self.phase_tick <= 0 and self.deathstate == 0:
                self.phase_down()
            if self.deathstate == 1:
                self.fadetimer -= self.g.delta_time
                self.deathoffset += vec2(600 * self.g.delta_time, 200 * self.g.delta_time)
                if self.fadetimer <= 0:
                    self.death()
        else:
            self.check_active()

    def phase_down(self):
        self.phase -= 1
        # fallback item
        for core in self.cores:
            core.death()
        for turret in self.turrets:
            turret.death()
        for eb in self.l.bulletsE:
            if isinstance(eb, Laser):
                eb.fadeout()
            else:
                eb.hit()
        if self.phase <= 0:
            self.deathstate = 1
            self.fadetimer = 5.5
            pygame.mixer.music.fadeout(1800)
        else:
            if self.phase!=3: self.floatoffset = 1000
            self.phase_change()

    def phase_change(self):
        for spawn in spawns[3 - self.phase]:
            item = self.l.create(spawn[0], self.pos)
            item.anchorpos = spawn[1]
            item.anchor = self
            if spawn[0] == "WhaleCore":
                self.cores.add(item)
            else:
                self.turrets.add(item)
                if spawn[0] == "WhaleTurretFlex" or spawn[0] == "WhaleTurretRocket":
                    item.shotcd = 5.5 + spawn[2]
                elif spawn[0] == "WhaleTurretSide":
                    item.shotcd = 5.5 + spawn[2]
                    item.rot = spawn[3]
                elif spawn[0] == "WhaleTurretLaser":
                    item.endpoints[0] = 420
                    item.endpoints[1] = -420
        if self.phase == 3:
            self._mhp = 250
        elif self.phase == 2:
            self._mhp = 500
            pass
        elif self.phase == 1:
            self._mhp = 500
            pass

    def death(self):
        self.l.bossDeathState = 1
        self.l.uncreate(self.instid)

    def render(self):
        if self.phase>0 and self._active:
            # Draw boss health bar at bottom of screen
            sz = self.g.screen_size
            mid = sz[0] / 2 - 490
            top = 80
            hel = float(self._hp / self._mhp) * 968
            pygame.draw.rect(self.l.s, (128,0,96,0)   , pygame.rect.Rect(mid  ,top  ,980,60))
            pygame.draw.rect(self.l.s, (255,0,192,100), pygame.rect.Rect(mid+6,top+6,hel,48))

        self.l.s.blit(self.image, pygame.rect.Rect(self.pos.x - 500 - 1000 * min(3-self.phase,2), self.pos.y - 292, 3000, 424).move(-self.l.view))


import pygame
import os

class RedPower(pygame.sprite.Sprite):
    def __init__(self, global_vars, level, pup_position, pup_width=32, pup_height=32):
        super().__init__()
        self.depth = 0
        self.instid = 4
        self.objid = 5
        self._global_vars = global_vars
        self._level = level
        self._pup_position = pup_position
        self.index = 0
        self.animation_counter = 0
        self.images = []
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images/Red")
        for frame in range(1,7):
            img = pygame.image.load(os.path.join(self.imageDir, f'frame {frame}.png')).convert_alpha()
            img = pygame.transform.scale(img, (48, 48))
            self.images.append(img)

        self.image = self.images[self.index]
        #self.rect = self.image.get_rect()
        self.rect = pygame.rect.Rect(pup_position[0], pup_position[1], pup_width, pup_height)
        #self.rect.center = self._pos
        self.player_acquired = False
        self._pup_width = pup_width


    def acquired(self):
        self.player_acquired = True
        return 3

    def update(self):
        self.animation_counter += 1
        if not self.player_acquired and self.animation_counter >= 10:
            self.image = self.images[self.index]
            self.index += 1
            self.animation_counter = 0
            if self.index >= 5:
                self.index = 0

    def render(self):
        spos = self._pup_position - self._level.view
        if not self.player_acquired:
            #pygame.draw.rect(self._level.s, (255, 215, 0, 100), self.rect.copy().move(-self._level.view))
            self._level.s.blit(self.image, self.rect.copy().move(-self._level.view))



from bluebeam.objects import Rocket, Explosion
import pygame, os, math


class HomingRocket(Rocket):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self.rect.width = self.rect.height = 48
        self.hp = 30
        self.lifespan = 10.
        self.lockon = 0.3
        self._dir = 270
        self._dirmax = 90
        self.maxvel = 275
        self.traveldir = pygame.math.Vector2(0,-1)

        self.sprites = []
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images/missile")
        self.load_sprites()

    def load_sprites(self):
        imgh = pygame.image.load(os.path.join(self.imageDir, f'missile1.png'))
        imgh = pygame.transform.scale(imgh, (48, 48))
        imgd = pygame.image.load(os.path.join(self.imageDir, f'missile9.png'))
        imgd = pygame.transform.scale(imgd, (80, 80))
        imgv = pygame.image.load(os.path.join(self.imageDir, f'missile18.png'))
        imgv = pygame.transform.scale(imgv, (48, 48))
        self.sprites.append(imgh)
        self.sprites.append(pygame.transform.flip(imgd,False,True))
        self.sprites.append(pygame.transform.flip(imgv,False,True))
        self.sprites.append(pygame.transform.flip(imgd,True,True))
        self.sprites.append(pygame.transform.flip(imgh,True,False))
        self.sprites.append(pygame.transform.flip(imgd,True,False))
        self.sprites.append(imgv)
        self.sprites.append(imgd)

    def update(self):
        self.lockon = max(0, self.lockon - self.g.delta_time)
        if self.lockon <= 0:
            self.aim()
        self.vel = self.traveldir * self.maxvel
        super().update()
        self.collision_bullet()
        pass

    def aim(self):
        ang = pygame.math.Vector2().angle_to(self.l.player.pos - self.pos)
        self._dir = ang
        self.traveldir = pygame.math.Vector2(math.cos(math.radians(self._dir)), math.sin(math.radians(self._dir)))

    def hit(self):
        boom = self.l.create("Explosion", self.pos.xy)
        boom.radius(192)
        self.l.uncreate(self.instid)

    def collision_bullet(self):
        col = pygame.sprite.spritecollide(self, self.l.bulletsP, False)
        if len(col) > 0:
            for bul in col:
                self.lifespan -= bul.dmg/2
                bul.hit()

    def render(self):
        num = int(((self._dir + 720) % 360) / 45) % 8
        self.l.s.blit(self.sprites[num], self.rect.copy().move(-self.l.view))


import math
import os
from bluebeam.objects import GameObject
import pygame


class GammaBossSpawnTrigger(GameObject):

    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self._activated = 0
        self.rect = pygame.rect.Rect(self.pos.x,self.pos.y,1024,256)
        self.rect.midtop = self.pos
        self.stand = []
        self.jump = []
        self.laugh = []
        self.clutch = []
        #paths so it works
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images")
        self.soundDir = os.path.join(os.path.dirname(self.sourceDir), "Sounds")
        #self.spritesheet = pygame.image.load(f'images/boss01_commander.png').convert()
        self.spritesheet = pygame.image.load(os.path.join(self.imageDir, "boss01_commander.png")).convert()
        self.load_sprites()
        self.commanderrp = pygame.math.Vector2(512, 0)
        self.waittospawn = 2.25
        self.waittojump = 0
        self.jumpS = self.commanderrp
        self.jumpE = self.commanderrp + pygame.math.Vector2(200,800)
        self.jumpForce = pygame.math.Vector2(50,-400)
        self.grav = 700
        self.waittospawn = 2.25
        self.waittojump = 0
        self.jumping = 0

    @property
    def activated(self):
        return self._activated

    @activated.setter
    def activated(self, a_):
        self._activated = a_

    def load_sprites(self):
        self.laugh.append(pygame.transform.scale(self.load_single((0, 0, 48, 48)), (144, 144)))
        self.laugh.append(pygame.transform.scale(self.load_single((48, 0, 48, 48)), (144, 144)))
        self.stand.append(pygame.transform.scale(self.load_single((0, 48, 48, 48)), (144, 144)))
        self.jump.append(pygame.transform.scale(self.load_single((48, 48, 48, 48)), (144, 144)))
        self.clutch.append(pygame.transform.scale(self.load_single((0, 96, 48, 48)), (144, 144)))
        self.image = self.stand[0]

    def firstframe(self):
        self.activated = 1
        self.l.view_target = self.pos + pygame.math.Vector2(-self.l.view_wh.x,-128 - self.l.view_wh.y)
        print(self.l.view)
        print(self.l.view_target)
        self.l.camstate = 1
        pygame.mixer.music.fadeout(1800)
        for eb in self.l.bulletsE:
            eb.hit()
        for e in self.l.enemies:
            e.hp = 0

    def update(self):
        self.image = self.stand[0]
        if self._activated == 1:
            if self.waittospawn > 0 and self.waittojump == 0:
                self.waittospawn = max(self.waittospawn - self.g.delta_time, 0)
                self.image = self.laugh[int(self.waittospawn * 7.5) % 2]
                if self.waittospawn == 0:
                    self.waittojump = 1.75
            elif self.waittojump > 0:
                self.waittojump -= self.g.delta_time
                self.image = self.jump[0]
                self.commanderrp += self.jumpForce * self.g.delta_time
                self.jumpForce += pygame.math.Vector2(0, self.grav) * self.g.delta_time
                if self.waittojump <= 0:
                    self.spawn()
                    self.l.uncreate(self.instid)

    def spawn(self):
        spire = self.l.create("GammaGenerator", self.pos)
        spire.setup([pygame.math.Vector2(self.rect.centerx-640,self.rect.top-512), pygame.math.Vector2(self.rect.centerx+640,self.rect.bottom)])
        pewpew0 = self.l.create("Laser", pygame.math.Vector2(self.rect.centerx-640, self.rect.top-768))
        pewpew0.dir = 1
        pewpew0.width = 32
        pewpew0.lifespan = math.pow(2,31) - 1
        pewpew0.warmup = 0
        pewpew0.times = 1
        pewpew1 = self.l.create("Laser", pygame.math.Vector2(self.rect.centerx+640, self.rect.top-768))
        pewpew1.dir = 1
        pewpew1.width = 32
        pewpew1.lifespan = math.pow(2,31) - 1
        pewpew1.warmup = 0
        pewpew1.times = 1
        self.waittojump = 0.8
        pygame.mixer.music.load(os.path.join(self.soundDir, 'db_ab.xm'))
        pygame.mixer.music.play(-1, 0, 200)

    def render(self):
        self.l.s.blit(self.image, self.rect.copy().move(-self.l.view + self.commanderrp + pygame.math.Vector2(0,self.rect.height-144)))


from bluebeam.objects import Enemy
import pygame, math
import os

class ShootyEnemy(Enemy):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self.rect = pygame.rect.Rect(0, 0, 108, 124)
        self.rect.center = self.pos
        self._mhp = 50
        self.hp = self._mhp
        self.idied = 0

        self.speed = pygame.math.Vector2(400, 800)

        self.shot_speed    = 400
        self.shot_cd       = 2
        self.panic_shot_cd = 0.5

        self.accel = 0.20
        self.gravity = 50
        self._shotcd = 0

        self.panic = False

        # for image/animation
        self.current_position = 0
        self.animation_counter = 0
        self.images_right = []
        self.images_left = []

        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images/Enemy")

        for image in range(1, 9):
            img_left = pygame.image.load(os.path.join(self.imageDir,f'shooter{image}.png'))
            img_left = pygame.transform.scale(img_left, (108, 124))
            img_right = pygame.transform.flip(img_left, True, False)
            self.images_right.append(img_right)
            self.images_left.append(img_left)

        self.image = self.images_left[self.current_position]

    def update_ai_inputs(self):
        self._rot = pygame.math.Vector2().angle_to(self.l.player.pos - self.pos)
        self.xmove()
        self.check_fire()

    def check_fire(self):
        self._shotcd = max(0, self._shotcd - self.g.delta_time)
        if (self._shotcd <= 0):
            self.fire()

    def xmove(self):
        if(self.l.player.x < self.x): 
            left = True
        else:
            left = False

        xspeed = self.speed.x * self.accel
        ydist = abs(self.l.player.y - self.y)
        xdist = abs(self.l.player.x - self.x)

        if((xdist > 700 or ydist > 700) and self.panic):
            self.panic = False
        elif((xdist <= 250 and ydist < 500) or self.panic):
            if(not left):  self.go_left(xspeed)
            else:     self.go_right(xspeed)

            self.try_jump()
            if(self.panic == False):
                self.panic = True
                self._shotcd = 0.3
        else:
            self.slow_down(xspeed)

        self.update_animation(left)

    def update_animation(self, left):
        # animation_counter = 0
        # if left:
        self.animation_counter += 1
        if self.animation_counter >= 5:
            self.animation_counter = 0
            self.current_position += 1
            if self.current_position >= 4:
                self.current_position = 0

        if left:
            self.image = self.images_left[self.current_position]
            # print(self.current_position)
        else:
            self.image = self.images_right[self.current_position]

    def go_left(self, xspeed):
        targ_xvel = -self.speed.x
        if (self.vel.x > targ_xvel):
            self.vel.x -= xspeed

    def go_right(self, xspeed):
        targ_xvel = self.speed.x
        if (self.vel.x < targ_xvel):
            self.vel.x += xspeed

    def slow_down(self, xspeed):
        if (self.vel.x < 0):
            self.vel.x += xspeed
        elif (self.vel.x > 0):
            self.vel.x -= xspeed

        if (abs(self.vel.x) < xspeed):
            self.vel.x = 0

    def try_jump(self):
        if (self.pstate == 0):
            self.jumping = True
            self._vel.y = -self.speed.y

    # This was Tyler's shooting mechanism for the boss.
    def fire(self):
        #bul = EnemyBullet("EnemyBullet", self.pos.xy)

        inst = self.l.create("EnemyBullet", self.pos.xy)

        inst.load_shooty_image()
        inst.collide_terrain = False
        inst.speed = pygame.math.Vector2(self.shot_speed * math.cos(math.radians(self._rot)),
                                         self.shot_speed * math.sin(math.radians(self._rot)))
        inst.shooty_enemy = True
        if self.l.player.x <= self.pos.x:
            inst.pos.x -= 100
            inst.pos.y -= 40
        else:
            inst.pos.x += 60
            inst.pos.y -= 40

        if (self.panic):
            self._shotcd = self.panic_shot_cd
        else:
            self._shotcd = self.shot_cd

    def update(self):
        super().update()
        if(self._active):
            self.update_ai_inputs()

    def render(self):
        self.l.s.blit(self.image, self.rect.copy().move(-self.l.view))
        pygame.draw.rect(self.l.s,
                         (255, 0, 0, 100),
                         pygame.rect.Rect(self.rect.topleft[0], self.rect.topleft[1] - 24, 100, 16).move(-self.l.view))
        pygame.draw.rect(self.l.s,
                         (0, 255, 0, 100),
                         pygame.rect.Rect(self.rect.topleft[0],
                                          self.rect.topleft[1] - 24,
                                          100 * (self.hp / self._mhp),
                                          16).move(-self.l.view))



from bluebeam.objects import EnemyBullet
import pygame, math


class SuperEnemyBullet(EnemyBullet):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self.lifespan = 10.
        self.targets = [pygame.math.Vector2(), pygame.math.Vector2()]
        self.specspeed = 400
        self.vel = pygame.math.Vector2(0,0)
        self.accel = 800
        self.targ = 0
        self.switchtarg = 0.8

    def update(self):
        cur = (self.targets[self.targ] - self.pos)
        if cur.magnitude() != 0 and self.vel != 0:
            self.vel = cur.normalize() * self.specspeed
        else:
            self.vel = pygame.math.Vector2(0)
        if (cur.magnitude() <= (self.vel * self.g.delta_time).magnitude()) and self.targ == 0:
            self.pos = self.targets[self.targ]
            self.vel = pygame.math.Vector2(0)
            self.switchtarg = max(0,self.switchtarg - self.g.delta_time)
            if self.switchtarg == 0:
                self.specspeed = pygame.math.Vector2(400)
                self.targ = 1
        elif self.targ == 1:
            self.specspeed += pygame.math.Vector2(self.accel * self.g.delta_time)
            self.vel = cur.normalize() * self.specspeed.x
        super().update()


import pygame
from bluebeam.objects import Bullet
import os


class PlayerBullet(Bullet):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        # self.owner = 0
        self.speed = pygame.math.Vector2(1000., 0.)
        self.lifespan = 5.
        self.dmg = 10
        self._objid = 4
        self.charged = False
        self.depth = 16
        self.rapid_fire = False
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images")
        self.image = pygame.image.load(os.path.join(self.imageDir, 'ball.png')).convert_alpha()
        self.image = pygame.transform.scale(self.image, (36, 36))
        self.large_image = pygame.transform.scale(self.image, (64, 64))

        self.collide_terrain = False


    def update(self):
        super().update()
        if(self.charged):
            self.collide_enemy_bullets()

    def collide_enemy_bullets(self):
        col = pygame.sprite.spritecollide(self, self.l.bulletsE, False)
        for bul in col:
            bul.hit()


    def render(self):
        spos = self._pos - self.l.view
        #pygame.draw.rect(self.l.s, (255, 255, 64, 100), self.rect.copy().move(-self.l.view))
        if not self.charged:
            self.l.s.blit(self.image, self.rect.copy().move(-self.l.view))
        else:
            self.l.s.blit(self.large_image, self.rect.copy().move(-self.l.view))




import math
import os

import pygame, itertools

from bluebeam.objects import *
from bluebeam.TileMap import TileMap


class Level:
    g = None  # global vars reference
    s = None  # screen reference
    # major objects
    view = pygame.math.Vector2()
    view_target = pygame.math.Vector2()
    player = pygame.sprite.Sprite()
    bg = None
    objtimer = 1
    camstate = 0
    # groups

    booster = pygame.sprite.Sprite()

    enemies = pygame.sprite.Group()
    pickups = pygame.sprite.Group()
    bulletsE = pygame.sprite.Group()  # it says "bullets" but melee attacks can just be wide, unmoving bullets :)
    bulletsP = pygame.sprite.Group()
    others = pygame.sprite.Group()
    powerUps = pygame.sprite.Group()
    triggers = pygame.sprite.Group()


    def __init__(self, g_, s_, cam_offset, FILENAME):
        self.view_wh = cam_offset
        self.g = g_
        self.s = s_
        self.terrain = pygame.sprite.Group()
        self.bossDeathState = 0
        self.load(FILENAME)
        self.player_around_lg = False

        pass

    def delete(self):
        objects = itertools.chain(self.bulletsP, self.bulletsE, self.enemies, self.powerUps, self.others, self.pickups, self.triggers)
        for thing in objects:
            thing.kill()
    
    def uncreate(self, id):
        temp = filter(lambda item: (item.instid == id),
                      itertools.chain(self.bulletsP, self.bulletsE, self.enemies, self.powerUps, self.triggers))
        for target in temp:
            target.kill()

    def load(self, FILENAME):
        # Background stuff
        self.TileMap = TileMap(self.g, self, filename=FILENAME)

        # Always load a player
        player_pos = self.TileMap.load_player_pos()
        self.player = Player(self.g, self, player_pos)

        # Enemies
        self.spawn_tilemap_enemies()
        self.spawn_tilemap_pickups()
        #self.create("AlphaWhale",pygame.math.Vector2(1000,300))
        #self.create("SliderEnemy",pygame.math.Vector2(3500,200))
        #self.create("LittleGrey", pygame.math.Vector2(100,2300))
        #self.enemies.add(self.create("LittleGrey", pygame.math.Vector2(100,2300)))

        #self.create("SliderEnemy",pygame.math.Vector2(1500,570))
        pygame.mixer.music.stop()

        sourceDir = os.path.dirname(os.path.abspath(__file__))
        soundDir = os.path.join(sourceDir, "Sounds")

        if FILENAME == "MapLevel2.tmx":
            pygame.mixer.music.load(os.path.join(soundDir, 'moskito.xm'))
        elif FILENAME == "MapLevel3.tmx":
            pygame.mixer.music.load(os.path.join(soundDir, 'heidi7.xm'))
        else:
            pygame.mixer.music.load(os.path.join(soundDir, 'flc_pamp.mod'))
        pygame.mixer.music.play(-1,0,0)
        pass

    def spawn_tilemap_pickups(self):
        pickups = self.TileMap.load_pickup_info()
        for picktype, pos in pickups:
            self.create(picktype, pos)

    def spawn_tilemap_enemies(self):
        enemy_list = self.TileMap.load_enemy_info()
        for enemy_type, enemy_pos in enemy_list:
            self.create(enemy_type, enemy_pos)

    def create(self, it, pos):
        item = eval(it)(self.g, self, pos)
        self.append(item)
        return item


    def append(self, item):
        if isinstance(item, Enemy):
            self.enemies.add(item)
        if isinstance(item, Bullet):
            if item.owner == 0:
                self.bulletsP.add(item)
            else:
                self.bulletsE.add(item)
        if isinstance(item, Block):
            self.terrain.add(item)

        if isinstance(item, BossSpawnTrigger) or isinstance(item, BetaBossSpawnTrigger) or isinstance(item, GammaBossSpawnTrigger):
            self.triggers.add(item)

        if isinstance(item, YellowPower):
             self.powerUps.add(item)

        if isinstance(item, BluePower):
            self.powerUps.add(item)
        if isinstance(item, RedPower):
            self.powerUps.add(item)

    def update(self):
        for bullet in itertools.chain(self.bulletsP, self.bulletsE):
            bullet.update()

        self.player.update()

        for enemy in self.enemies:
            enemy.update()

        for powerUp in self.powerUps:
            powerUp.update()

        for trigger in self.triggers:
            trigger.update()

        self.check_grey()

        self.update_view()
        # bg.update()
        self.objtimer -= self.g.delta_time
        if self.objtimer <= 0:
            self.objtimer = 1.
            num = len(self.bulletsE) + len(self.bulletsP) + len(self.enemies) + len(self.pickups) + len(
                self.others) + len(self.powerUps)
        pass

    def update_view(self):
        if self.camstate == 0:
            self.view = self.player.pos - self.view_wh
        elif self.camstate == 1:
            dist = (self.view_target - self.view).length()
            if dist > 0:
                ang = (self.view_target - self.view).normalize()
                self.view += ang * min(dist, 600 * self.g.delta_time)
        # print("self.player.pos: ", self.player.pos)

    def check_grey(self):
        found_grey = False
        for enemy in self.enemies:
            if enemy._telepath:
                self.player.around_lg = True
                self.player.confuse_cd = 1.5
                found_grey = True
        if(not found_grey):
            if(self.player.confuse_cd > 0):
                self.player.confuse_cd -= self.g.delta_time
            if(self.player.confuse_cd <= 0):
                self.player.around_lg = False
        

    def render(self):
        self.s.fill((0, 0, 255))
        self.TileMap.render(self.s, self.view)
        self.renderpipeline = pygame.sprite.Group(sorted(itertools.chain(self.enemies, self.bulletsE, self.bulletsP, self.triggers, self.powerUps), key=lambda d: d.depth))
        for item in self.renderpipeline:
            item.render()
        #for enemy in self.enemies:
        #    enemy.render()
        #for bullet in itertools.chain(self.bulletsE, self.bulletsP):
        #    bullet.render()
        self.player.render()
        #for power_up in self.powerUps:
        #    power_up.render()
        #for trigger in self.triggers:
        #    trigger.render()
        # for pickup in pickups:
        #     pickup.render()
        # for misc in others:
        #   misc.render()
        pass



from bluebeam.objects import WhaleTurret
import pygame


class WhaleShield(WhaleTurret):

    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self.hp = 100
        self.rect = pygame.rect.Rect(self.pos.x, self.pos.y, 160, 32)
        self.rect.center = self.pos

    def render(self):
        pygame.draw.rect(self.l.s, (192, 32, 32, 100), self.rect.copy().move(-self.l.view))



from bluebeam.objects import EnemyBullet
import pygame


class Shockwave(EnemyBullet):

    def __init__(self, g_, l_, pos_):
        super().__init__(g_, l_, pos_)
        self._frames = 10
        self._frame = 10
        self._spawnat = self._frames - 1
        self._timeperframe = 0.02
        self._width = 16
        self._height = 256
        self.depth = 16
        self._rheight = self._height * (1 - (1 + abs(self._frame - self._frames / 2)) / (1 + self._frames / 2))
        self.rect = pygame.rect.Rect(self.pos.x, self.pos.y, self._width, self._rheight)
        self.rect.midbottom = self.pos
        self.lifespan = self._timeperframe
        self.dir = 1
        self.dmg = 10

    def setsize(self, v2_):
        self._width = v2_.x
        self._height = v2_.y
        self._rheight = self._height * (1 - (1 + abs(self._frame - self._frames / 2)) / (1 + self._frames / 2))
        self.rect = pygame.rect.Rect(self.pos.x, self.pos.y, v2_.x, self._rheight)
        self.rect.midbottom = self.pos

    def update(self):
        self._rheight = self._height * (1 - (abs(self._frame - self._frames / 2)) / (1 + self._frames / 2))
        self.rect = pygame.rect.Rect(self.pos.x, self.pos.y, self._width, self._rheight)
        self.rect.midbottom = self.pos
        self._lifespan = max(0, self._lifespan - self.g.delta_time)
        if self._lifespan <= 0:
            if self._frame == -1:
                self.fadeout()
            else:
                self._lifespan = self._timeperframe
                self._frame -= 1
                if self._frame == self._spawnat:
                    if not self.check(self._width * self.dir):
                        nextwave = self.l.create("Shockwave", self.pos + pygame.math.Vector2(self._width * self.dir, 0))
                        nextwave.setsize(pygame.math.Vector2(self._width, self._height))
                        nextwave.dir = self.dir
                        nextwave.setframes(self._frames, self._timeperframe)

    def fadeout(self):
        self.l.uncreate(self.instid)

    def hit(self):
        pass

    def setframes(self, f_, t_=None):
        if t_ is not None:
            self._timeperframe = t_
        self._frames = f_
        self._frame = f_
        self._spawnat = f_ - 1

    def check(self, offset):
        print(self.rect.height)
        self.rect = pygame.rect.Rect(self.pos.x, self.pos.y, self._width, self._rheight)
        self.rect.midbottom = self.pos
        col = pygame.sprite.spritecollide(self, self.l.terrain, False)
        if len(col) > 0:
            return 1
        else:
            return 0



from bluebeam.objects import WhaleTurret
import pygame


class WhaleTurretLaser(WhaleTurret):
    def __init__(self, g_, l_, pos_=None):
        self.style = 3
        super().__init__(g_, l_, pos_)
        self._hp = 150
        self._endpoints = [-100, 100]
        self._railx = 0
        self.dir = 1
        self._emitpoint = pygame.math.Vector2(self.pos.x, (self.pos.y + self.rect.midbottom[1]) / 2)
        self._shotcd = 4.67
        self._rot = 1

    @property
    def endpoints(self):
        return self._endpoints

    @endpoints.setter
    def endpoints(self, e_):
        self._endpoints = e_

    @property
    def emitpoint(self):
        return self._emitpoint

    @emitpoint.setter
    def emitpoint(self, e_):
        self._emitpoint = e_

    def update(self):
        super().update()
        self._railx += self.dir * 160 * self.g.delta_time
        if self.dir * self._railx > self.dir * self.endpoints[int(0.5 - 0.5 * self.dir)]:
            self._railx = self.endpoints[int(0.5 - 0.5 * self.dir)]
            self.dir = -self.dir
        self.rect.center = self.pos = self._anchor.pos + self._anchorpos + pygame.math.Vector2(self._railx, 0)
        self._emitpoint = pygame.math.Vector2(self.pos.x, (self.pos.y + self.rect.midbottom[1]) / 2)

    def fire(self):
        inst = self.l.create("Laser", self.pos.xy)
        inst.emitter = self
        inst.dir = self._rot
        inst.width = 36
        self._shotcd = 4.67

    def render(self):
        self.rot = self.rot * 90
        super().render()
        self.rot = self.rot / 90



import pygame


class Background:
    _g = None
    _l = None

    layer_img = []
    layer_spd = []  # speeds for parallax scrolling
    layer_pos = []

    def __init__(self, g_, l_):
        self.layer()
        self._g = g_
        self._l = l_

    def load(self):
        pass

    def layer(self):
        self.layer_img.append(pygame.image.load("images/test_background.png").convert())
        # self.layer_img[0].fill((128, 0, 0, 100), special_flags=pygame.BLEND_ADD)
        self.layer_spd.append(0.)
        self.layer_pos.append(pygame.math.Vector2(-700, -700))

    def update(self):
        # parallax scrolling logic will go here c:
        pass

    def render(self):
        for i in range(0, len(self.layer_img)):
            spos = self.layer_pos[i] - self._l.view
            self._l.s.blit(self.layer_img[i], spos)



import pygame, math
from bluebeam.objects import EnemyBullet
from pygame import Vector2 as vec2


class WaveBomb(EnemyBullet):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self._initvv = 0
        self.gravity = 35
        self.vel = pygame.math.Vector2(0, self._initvv)
        self.rect = pygame.rect.Rect(self.pos.x, self.pos.y, 64, 64)
        self.rect.center = self.pos
        self.depth = 16

    @property
    def splits(self):
        return self._splits

    @splits.setter
    def splits(self, s_):
        self._splits = s_

    def update(self):
        self.update_phys()

    def hit(self):
        pass

    def split(self):
        pew = self.l.create("Shockwave", self.rect.midbottom + vec2(0,-1))
        pew.setsize(vec2(32,256))
        pew.setframes(2, 0.03)
        pew.dir = 1
        pew2 = self.l.create("Shockwave", self.rect.midbottom + vec2(0,-1))
        pew2.setsize(vec2(32,256))
        pew2.setframes(2, 0.03)
        pew2.dir = -1
        self.l.uncreate(self.instid)

    def ccheck_h(self):
        cc = self.rect.copy()
        for i in range(1, math.ceil(abs(self.vel.x))):
            self.rect = self.rect.move(math.copysign(1, self.vel.x), 0)
            if self.collision_terrain():
                self.vel = pygame.math.Vector2(-self.vel.x, self.vel.y)
                break
        self.rect = cc

    def ccheck_v(self):
        if (self.pstate == 0): self.pstate = 1
        cc = self.rect.copy()
        for i in range(1, math.ceil(abs(self.vel.y))):
            self.rect = self.rect.move(0, math.copysign(1, self.vel.y))
            if self.collision_terrain(True):
                self.rect = self.rect.move(0, -1)
                self.split()
                break
        self.rect = cc

    def collision_terrain(self, vertical=False):
        col = pygame.sprite.spritecollide(self, self.l.terrain, False)
        if len(col) > 0:
            return 1
        else:
            return 0



from bluebeam.objects import Enemy
from bluebeam.objects import Tentacle
import pygame
import os


class BulkyEnemy(Enemy):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self.rect = pygame.rect.Rect(0, 0, 150, 245)
        self.rect.center
        self._mhp = 150
        self.hp = self._mhp
        self.idied = 0

        self.speed = pygame.math.Vector2(550, 1100)
        self.accel = 0.05
        self.gravity = 80
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images/Enemy/BulkyGuy")

        self.max_attack_cd = 1
        self.attack_cd = self.max_attack_cd
        self.tentacle = None

        self.left = True

        #for image/animation
        self.attacking = False
        self.current_position = 0
        self.animation_counter = 0
        self.walk_right = []
        self.walk_left = []
        for image in range(1,4):
            img_right = pygame.image.load(os.path.join(self.imageDir,f'BulkWalk{image}.png')).convert_alpha()
            img_right = pygame.transform.scale(img_right, (326, 258))
            img_left = pygame.transform.flip(img_right, True, False)
            self.walk_right.append(img_right)
            self.walk_left.append(img_left)

        self.attack_right = []
        self.attack_left = []

        for image in range(1,8):
            img_right = pygame.image.load(os.path.join(self.imageDir,f'BulkAttack{image}.png')).convert_alpha()
            img_right = pygame.transform.scale(img_right, (326, 258))
            img_left = pygame.transform.flip(img_right, True, False)
            self.attack_right.append(img_right)
            self.attack_left.append(img_left)



        self.image = self.walk_left[self.current_position]


    def update_ai_inputs(self):
        if( self.attacking == False): 
            self.try_attack()
        self.xmove()

    def try_attack(self):
        if(self.attack_cd <= 0):
            above = self.l.player.y < self.y - 100
            xdist = abs(self.l.player.x - (self.x + self.vel.x * self.g.delta_time))
            if(xdist <= 400):
                if(above):
                    self.try_jump(cd=1)
                else:
                    self.attacking = True
                    self.animation_counter = 0
                    self.current_position = 0
            elif(above):
                self.try_jump(cd=0.3)
            elif(xdist >= 600 and self.l.player.y < self.y - 50):
                self.try_jump(cd=0)
        else:
            self.attack_cd -= self.g.delta_time

    def xmove(self):
        # Should I go left or right?
        if(self.attacking == False or self.current_position < 2):
            if (self.l.player.x < self.x - 80):
                targ_xvel = -self.speed.x
                if (self.vel.x > targ_xvel):
                    self.vel.x -= self.speed.x * self.accel

            elif (self.l.player.x > self.x + 80):
                targ_xvel = self.speed.x
                if (self.vel.x < targ_xvel):
                    self.vel.x += self.speed.x * self.accel
                    #left = False
            if (self.l.player.x < self.x):
                self.left = True
            else:
                self.left = False
        else:
            self.slow_down(self.speed.x * .01)
        self.update_animation(self.left)
    
    def try_jump(self, cd=0.3):
        if (self.pstate == 0):
            if(self.attack_cd <= 0):
                self.jumping=True
                self._vel.y = -self.speed.y
                self.attack_cd = cd

    def slow_down(self, xspeed):
        if (self.vel.x < 0):
            self.vel.x += xspeed
        elif (self.vel.x > 0):
            self.vel.x -= xspeed

        if (abs(self.vel.x) < xspeed):
            self.vel.x = 0

    def update_animation(self, left):
        attack_delays = [15,25,20,3,3,20,5]
        #attack_delays = [50,50,50,50,50,50,50]
        #animation_counter = 0
        #if left:
        self.animation_counter += 1
        if self.attacking:
            if self.animation_counter >= attack_delays[self.current_position]:
                if(self.tentacle):
                    self.l.uncreate(self.tentacle.instid)
                    self.tentacle = None
                self.animation_counter = 0
                self.current_position += 1
                tentacle_num = self.current_position + 1
                if (4 <= tentacle_num and tentacle_num <= 7):
                    self.tentacle = Tentacle.Tentacle(self.g, self.l, tentacle_num, self.left, self._pos)
                    self.tentacle._from_bulky = True
                    self.l.bulletsE.add(self.tentacle)
                if self.current_position >= 7:
                    self.current_position = 0
                    self.attacking = False
                    self.attack_cd = self.max_attack_cd
            
        else:
            if self.animation_counter >= 10:
                self.animation_counter = 0
                self.current_position += 1
                if self.current_position >= 3:
                    self.current_position = 0

        if self.attacking:
            if left: self.image = self.attack_left[self.current_position]
            else: self.image = self.attack_right[self.current_position]
        else:
            if left: self.image = self.walk_left[self.current_position]
            else: self.image = self.walk_right[self.current_position]

    def death(self):
        if(self.tentacle):
            self.l.uncreate(self.tentacle.instid)
        self.l.uncreate(self.instid)

    def update(self):
        super().update()
        if(self._active):
            self.update_ai_inputs()
        if(self.tentacle):
            self.tentacle.refresh_pos(self._pos + self._vel*self.g.delta_time)

    def render(self):
        # pygame.draw.rect(self.l.s,
        #                  (128 + 127 * (1 - self.idied), 128, 128 + (127 * self.idied), 100),
        #                  self.rect.copy().move(-self.l.view))
        
        if(self.left):
            self.l.s.blit(self.image, self.rect.copy().move(-self.l.view - (150,0)))
        else:
            self.l.s.blit(self.image, self.rect.copy().move(-self.l.view - (0,0)))
        pygame.draw.rect(self.l.s,
                         (255, 0, 0, 100),
                         pygame.rect.Rect(self.rect.topleft[0], self.rect.topleft[1] - 24, 100, 16).move(-self.l.view))
        pygame.draw.rect(self.l.s,
                         (0, 255, 0, 100),
                         pygame.rect.Rect(self.rect.topleft[0],
                                          self.rect.topleft[1] - 24,
                                          100 * (self.hp / self._mhp),
                                          16).move(-self.l.view))


from bluebeam.objects import Target
import pygame


class BadTarget(Target):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self.shotcd = 1.67

    def update(self):
        super().update()
        if (self.shotcd <= 0):
            inst = self.l.create("EnemyBullet", self.pos.xy)
            inst.speed = pygame.math.Vector2(-300, 0)
            self.shotcd = 1.67
        else:
            self.shotcd -= self.g.delta_time



import pygame
import os
from bluebeam.objects import Bullet



class EnemyBullet(Bullet):
    def __init__(self, g_, l_, pos_=None, left=False, above=False):
        super().__init__(g_, l_, pos_)
        self.owner = 1
        self.speed = pygame.math.Vector2(800., 0.)
        self.vel = self.speed
        self.lifespan = self.l.view_wh.length() / 300
        self._objid = 4
        self.shooty_enemy = False
        self.from_bulky = False

        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images")

        if( not hasattr(self, "image") ):
            self.load_ball_img()

    def load_ball_img(self):
        self.image = pygame.image.load(os.path.join(self.imageDir, 'ball_red.png')).convert_alpha()
        self.image = pygame.transform.scale(self.image, (64, 64))



    def load_shooty_image(self):
        self.images = []
        self.index = 0
        #getting paths so it works after pip install
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images/stars")
        for frame in range(1, 4):
            #img = pygame.image.load(f'images/stars/s{frame}.png')
            img = pygame.image.load(os.path.join(self.imageDir, f's{frame}.png'))
            img = pygame.transform.scale(img, (80, 80))
            self.images.append(img)

        self.image = self.images[self.index]
        self.current_position = 0
        self.animation_counter = 0


    def update(self):
        super().update()
        if self.shooty_enemy:
            self.update_animation()

    def update_animation(self):
        self.animation_counter += 1
        if self.animation_counter >= 5:
            self.image = self.images[self.index]
            self.index += 1
            self.animation_counter = 0
            if self.index >= 3:
                self.index = 0


        #self.l.s.blit(self.image, self.rect.copy().move(-self.l.view))
        #print("Is this not being called?")

    def render(self):
        spos = self._pos - self.l.view
        if self.shooty_enemy:
            self.l.s.blit(self.image, self.rect.copy().move(-self.l.view))
        elif not self.from_bulky:
            #dpygame.draw.rect(self.l.s, (255, 64, 64, 100), self.rect.copy().move(-self.l.view))
            self.l.s.blit(self.image, self.rect.copy().move(-self.l.view))




import pygame


class Block(pygame.sprite.Sprite):
    def __init__(self, global_vars, level, block_position, block_width=64, block_height=64):
        super().__init__()
        self._global_vars = global_vars
        self._level = level
        self._block_position = block_position
        self.rect = pygame.rect.Rect(block_position[0], block_position[1], block_width, block_height)
        self.block_width = block_width

    def render(self):
        spos = self._block_position - self._level.view
        pygame.draw.rect(self._level.s, (255, 255, 255, 100), self.rect.copy().move(-self._level.view))



from bluebeam.objects import EnemyBullet
import pygame
import os


class Explosion(EnemyBullet):

    def __init__(self, g_, l_, pos_):
        super().__init__(g_, l_, pos_)
        self.lifespan = 15
        self.dmg = 20

        self.images = []
        self.index = 0
        self.animation_counter = 0
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images/Explosion")
        for frame in range(1, 5):
            img = pygame.image.load(os.path.join(self.imageDir, f'explosion{frame}.png')).convert_alpha()
            img = pygame.transform.scale(img, (210, 210))
            self.images.append(img)

        self.image = self.images[self.index]

        pass

    def radius(self, radi):
        self.rect = pygame.rect.Rect(self.pos.x - radi / 2, self.pos.y - radi / 2, radi, radi)
        self.rect.center = self.pos

    def hit(self):
        pass

    def animate(self):
        self.animation_counter += 1
        self.lifespan -= 1
        if self.animation_counter >= 3:
            self.image = self.images[self.index]
            self.index += 1
            self.animation_counter = 0
            if self.index >= 4:
                self.lifespan - 0
                self.l.uncreate(self.instid)

    def update(self):
        self.animate()
        '''
        if self.lifespan <= 0:
            self.l.uncreate(self.instid)
        else:
            self.lifespan -= 1
        '''

    def render(self):
        #pygame.draw.circle(self.l.s, (255, 255, 128, 100), self.rect.center - self.l.view, self.rect.width / 1.71)
        self.l.s.blit(self.image, self.rect.copy().move(-self.l.view))



#GAMMAGENERATOR - final boss
from bluebeam.objects import Enemy
import pygame, os, random
from pygame import Vector2 as vec2


class GammaGenerator(Enemy):
    def __init__(self, g, l, pos):
        super().__init__(g, l, pos)
        self.rect = pygame.rect.Rect(0, 0, 832, 768)
        self.phase = 3
        self.hp = 1
        self._mhp = 1

        self.depth = 8
        self.bob = 0

        self.bounds = [vec2(), vec2()]
        self.groundlevel = 0
        self.hands = []
        self.head = None
        self.rocketpoints = [vec2(), vec2(), vec2(), vec2()]
        self.core = None

        self.body = None
        self.rocket = None
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imagedir = os.path.join(os.path.dirname(self.sourceDir), "images/Boss/GammaGenerator")
        self.spritesheet = pygame.image.load(os.path.join(self.imagedir, 'boss03_body.png')).convert()
        self.load_sprites()

        self.orders = 3
        self.ordertimer = 0
        self.supermove = 0

        self.attackid = 0
        self.attackamt = 0
        self.attacktimer = 0

        self.explo = 0
        self.explo_time = 0

        self.deathstate = 0
        self.risetimer = 3
        self.fadetimer = 0

    def load_sprites(self):
        img = self.load_single((0,0,208,192))
        self.body = pygame.transform.scale(img, (832,768))
        self.spritesheet = pygame.image.load(os.path.join(self.imagedir, 'boss03_rockets.png')).convert()
        img = self.load_single((0,0,42,69))
        self.rocket = pygame.transform.scale(img, (168,276))

    def setup(self, bounds_):
        self.bounds = bounds_
        self.rect.center = self.pos = vec2((self.bounds[0].x + self.bounds[1].x) / 2, self.bounds[1].y + 560)
        tpos = vec2((self.bounds[0].x + self.bounds[1].x) / 2, self.bounds[1].y - 360)
        self.groundlevel = self.bounds[1].y
        for i in range(0,4):
            x = self.pos.x + (2*(i%2)-1)*(self.rect.width/2)
            y = self.pos.y - 64 + (i>=2)*64
            hand = self.l.create("GammaHand",vec2(x,y))
            hand.setup(self, i % 2, i >= 2)
            self.hands.append(hand)
            x = tpos.x + 240 - 480*(i%2) + (40 - 80*(i%2)) * int(i/2)
            y = tpos.y + 160 * int(i/2)
            self.rocketpoints[i] = vec2(x,y)
        self.head = self.l.create("GammaHead", vec2(self.pos.x, self.pos.y - 256))
        self.head.setup(self)
        self.core = self.l.create("GammaCore", self.pos.xy)
        self.core.setup(self)
        self.depth = 9
        self.ordertimer = 3
        self.attackamt = 0
        self.phase_change()

    def update(self):
        if self.explo > 0:
            self.phase_explo()
        if self.deathstate == 0:
            if self.risetimer > 0:
                s = vec2((self.bounds[0].x + self.bounds[1].x) / 2, self.bounds[1].y + 560)
                c = vec2((self.bounds[0].x + self.bounds[1].x) / 2, self.bounds[1].y - 360)
                self.risetimer = max(0,self.risetimer-self.g.delta_time)
                self.pos = self.rect.center = c.lerp(s, self.risetimer/3)
                self.ordertimer = 1.75
            self.bob = (self.bob + self.g.delta_time * 180) % 360
            if self.attacktimer > 0:
                self.attacktimer = max(self.attacktimer - self.g.delta_time, 0)
            elif self.ordertimer > 0 and self.attackamt <= 0:
                self.ordertimer = max(self.ordertimer - self.g.delta_time, 0)
            if self.ordertimer <= 0 and self.attackamt <= 0 and self.attacktimer <= 0:
                if self.supermove == 1:
                    self.superorder()
                else:
                    self.order()
            elif self.attackamt > 0 and self.attacktimer <= 0:
                self.attack()
            if self.hp <= 0:
                self.phase_down()
        elif self.deathstate == 1:
            self.fadetimer = max(self.fadetimer - self.g.delta_time,0)
            if self.fadetimer < 4.5:
                self.pos += vec2(0,200) * self.g.delta_time
                self.rect.center = self.pos
            if self.fadetimer == 0:
                self.finaldeath()

    def order(self):
        if self.orders > 0:
            neworder = random.randint(0, 4)
            self.ordertimer = 1
            if neworder==0:
                # attack 0 - fist swing
                self.attackid = 0
                self.attackamt = 4
                self.attack()
                pass
            elif neworder==1:
                # attack 1 - radial barrage
                self.attackid = 1
                self.attackamt = 1
                self.attack()
                pass
            elif neworder==2:
                self.attackid = 2
                self.attackamt = 1
                self.attack()
                # attack 2 - shockwave bombs
                pass
            elif neworder==3:
                # attack 3 - background missiles
                self.attackid = 3
                self.attackamt = 4 + 2*(3-self.phase)
                self.attack()
                pass
            elif neworder==4:
                # attack 4 - rapid lasers
                self.attackid = 4
                self.attackamt = 2 + (3-self.phase)
                self.attack()
                pass
            else: self.superorder()
        else:
            self.orders = 3 + (3-self.phase)  # 3 orders on phase 1 + an order for each later phase
            self.core.exposed = 6.5
            self.ordertimer = self.core.exposed + 3
        pass

    def attack(self):
        if self.attackid == 0:
            if self.phase==3:
                self.hands[self.attackamt % 2].state(1)
            else:
                self.hands[self.attackamt % 2].state(1)
                self.hands[2 + self.attackamt % 2].state(1)
            self.attacktimer = 1 + (self.attackamt%2 == 1)*1
            self.attackamt -= 1
            pass
        elif self.attackid == 1:
            for i in range(0,2):
                self.hands[i].state(2)
            self.attacktimer = 5.25 + 1.3*(3-self.phase)
            self.attackamt -= 1
            pass
        elif self.attackid == 2:
            self.head.state(1)
            self.attacktimer = 5.25
            self.attackamt -= 1
            pass
        elif self.attackid == 3:
            i = (4 - self.attackamt) % 4
            rock = self.l.create("GammaBGRocket", self.l.player.pos.xy)
            rock.setup(self.rocketpoints[i])
            self.attackamt -= 1
            self.attacktimer = 0.85 - 0.12*(3-self.phase)
            pass
        elif self.attackid == 4:
            i = (2 - self.attackamt) % 2
            self.hands[2+i].state(3)
            self.attackamt -= 1
            self.attacktimer = (self.bounds[1].x - self.bounds[0].x) / (400 + 30*(3-self.phase)) + 0.7
            pass
        else:
            self.superattack()
        if self.attackamt <= 0:
            self.ordertimer = 2.5
            self.orders -= 1

    def superattack(self):
        for i in range(0,4):
            self.hands[i].state(5 + int(i/2))
            self.hands[3].corridorclock = 0.5
            self.attacktimer = 15
            self.attackamt -= 1
            self.ordertimer = 3
            self.supermove = 0

    def superorder(self):
        # superorders - used when phase 2 and 3 start
        self.attackid = 666
        self.attackamt = 1
        self.superattack()
        # corridor attack

    def spawnhealth(self):
        l = (self.bounds[1].x - self.bounds[0].x) / 4
        for i in range(0, 4):
            x = self.bounds[0].x + l*(i+0.5)
            y = self.bounds[1].y - 96
            node = self.l.create("RedPower",vec2(x,y))
            node.depth = 9

    def phase_down(self):
        self.spawnhealth()
        self.phase -= 1

        if self.phase <= 0:
            self.deathstate = 1
            self.fadetimer = 5.5
            for i in range(0,4):
                self.hands[i].death()
            pygame.mixer.music.fadeout(1800)
            self.explo = 80
        else:
            for i in range(0, 3):
                self.hands[i].state(0)
            self.phase_change()
            self.core.exposed = 0
            self.ordertimer = 5
            self.supermove = 1
            self.explo = 16

    def phase_explo(self):
        self.explo_time -= self.g.delta_time
        if self.explo_time <= 0:
            self.explo -= 1
            self.explo_time = 0.1
            boom = self.l.create("Explosion", self.pos.xy + vec2(random.randint(-300,300), random.randint(-300,300)))
            boom.dmg = 0
            boom.radius(160)
            boom.depth = 64
        pass

    def phase_change(self):
        if self.phase == 3:
            self._mhp = 100
        elif self.phase == 2:
            self._mhp = 110
        elif self.phase == 1:
            self._mhp = 120
        self.hp = self._mhp

    def death(self):
        pass

    def finaldeath(self):
        self.l.bossDeathState = 1
        bigboom = self.l.create("Explosion",self.pos.xy)
        bigboom.dmg = 0
        bigboom.radius(2048)
        bigboom.depth = 64
        self.l.uncreate(self.instid)

    def render(self):
        self.l.s.blit(self.body,self.rect.copy().move(-self.l.view))
        self.l.s.blit(self.rocket, self.rect.copy().move(-self.l.view + vec2(60,300)))
        self.l.s.blit(pygame.transform.flip(self.rocket,True,False), self.rect.copy().move(-self.l.view + vec2(612,300)))
        if self.phase > 0:
            # Draw boss health bar at bottom of screen
            sz = self.g.screen_size
            mid = sz[0] / 2 - 490
            top = 80
            hel = float(self._hp / self._mhp) * 968
            pygame.draw.rect(self.l.s, (128, 0, 96, 0), pygame.rect.Rect(mid, top, 980, 60))
            pygame.draw.rect(self.l.s, (255, 0, 192, 100), pygame.rect.Rect(mid + 6, top + 6, hel, 48))


import pygame


class GameObject(pygame.sprite.Sprite):
    g = None

    # Initialize object, call a creation function

    def __init__(self, g_, l_, pos_=None):
        super().__init__()
        self.g = g_  # ref to global vars
        self.l = l_  # ref to level
        self._objid = 0  # CONST!
        self._instid = self.g.getid()  # Gets instance id from global vars
        self.spritesheet = None
        self.depth = 0
        if isinstance(pos_, pygame.math.Vector2):
            self._pos = pos_
        else:
            self._pos = pygame.math.Vector2(0, 0)
        # self.image = pygame.image.load("MaybeWeWontNeedToUseThisButICommentedItHereAny.way")
        # self.rect = self.image.get_rect()
        self.rect = pygame.rect.Rect(0, 0, 0,
                                     0)  # Changing values 0 and 1 here are irrelevant. 2 is width. 3 is height.
        self.rect.center = self.pos
        self.depth = 0
        self._flag = (1, 1, 0)
        self.create()

    @property
    def objid(self):
        return self._objid

    @property
    def instid(self):
        return self._instid

    @property
    def pos(self):
        return self._pos

    @pos.setter
    def pos(self, n_pos):
        self._pos = n_pos

    @property
    def flag(self):
        return self._flag

    @flag.setter
    def flag(self, n_flag):
        self._flag = n_flag

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, n_mask):
        self._mask = n_mask

    @property
    def x(self):
        return self.pos[0]

    @property
    def y(self):
        return self.pos[1]

    def load_single(self, vec4):
        rect = pygame.Rect(vec4[0], vec4[1], vec4[2], vec4[3])
        img = pygame.Surface(rect.size).convert()
        img.blit(self.spritesheet, (0, 0), rect)
        img.set_colorkey(img.get_at((0, 0)), pygame.RLEACCEL)
        return img

    def create(self):
        pass

    def update(self):
        pass

    def delete(self):
        pass

    # generic render function
    def render(self):
        # nothing
        pass



from bluebeam.objects import Enemy
import pygame
import math
import os

# Base class for whale boss turrets


class WhaleTurret(Enemy):

    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self._hp = 50
        self._contactdmg = 0
        self._anchor = None
        self._anchorpos = pygame.math.Vector2(0, 0)
        self._dmgflash = 0
        self._shotcd = 0.01
        self._rot = 0
        self.rect = pygame.rect.Rect(self.pos.x - 25, self.pos.y - 25, 80, 80)
        self.rect.center = self.pos
        self.osc = 0
        self.bases = []
        self.guns = []
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images")
        self.spritesheet = pygame.image.load(os.path.join(self.imageDir, 'boss01_turrets.png')).convert()
        if not hasattr(self, "style"): self.style = 0
        self.load_sprites()

    @property
    def anchor(self):
        return self._anchor

    @anchor.setter
    def anchor(self, a_):
        self._anchor = a_

    @property
    def anchorpos(self):
        return self._anchorpos

    @anchorpos.setter
    def anchorpos(self, ap_):
        self._anchorpos = ap_

    @property
    def shotcd(self):
        return self._shotcd

    @shotcd.setter
    def shotcd(self, sc_):
        self._shotcd = sc_

    @property
    def rot(self):
        return self._rot

    @rot.setter
    def rot(self, r_):
        self._rot = r_

    def load_single(self, vec4):
        rect = pygame.Rect(vec4[0], vec4[1], vec4[2], vec4[3])
        img = pygame.Surface(rect.size).convert()
        img.blit(self.spritesheet, (0, 0), rect)
        img.set_colorkey(img.get_at((0, 0)), pygame.RLEACCEL)
        return img

    def load_sprites(self):
        for image in range(4):
            x = 88 * (image % 4)
            self.bases.append(self.load_single((x,0,88,88)))
            self.guns.append(self.load_single((x,88,88,88)))
        self.image = self.bases[self.style]

    def update(self):
        super().update()
        self.pos = self._anchor.pos + self._anchorpos
        self.vel = 0
        self._dmgflash = max(0, self._dmgflash - self.g.delta_time)
        self._shotcd = max(0, self._shotcd - self.g.delta_time)
        if self._shotcd <= 0:
            self.fire()
        self.image = self.bases[self.style]

    def fire(self):
        self._shotcd = 2.5

    def collision_bullet(self):
        col = pygame.sprite.spritecollide(self, self.l.bulletsP, False)
        if len(col) > 0:
            for bul in col:
                self.hp -= bul.dmg
                self.l.uncreate(bul.instid)
                self._dmgflash = 0.35

    def rotinplace(self, image, angle):
        core = image.get_rect().center
        ret = pygame.transform.rotate(image, angle)
        rec = ret.get_rect(center = ret.get_rect(center = (core[0],core[1])).center)
        return ret, rec

    def render(self):
        # Draw base
        self.l.s.blit(self.image, self.rect.copy().move(-self.l.view))
        # Draw gun
        gun = self.guns[self.style]
        rgun = self.rotinplace(gun, 90 + -self.rot)
        self.l.s.blit(rgun[0], rgun[1].copy().move(-self.l.view + self.pos - pygame.math.Vector2(self.rect.width/2,self.rect.height/2)))


from bluebeam.objects import Enemy
import pygame
import math
import os

class WhaleCore(Enemy):

    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self._hp = 250
        self._contactdmg = 0
        self._anchor = None
        self._anchorpos = pygame.math.Vector2(0, 0)
        self._dmgflash = 0
        self.rect = pygame.rect.Rect(self.pos.x - 75, self.pos.y - 75, 100, 100)
        self.rect.center = self.pos
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images")
        self.spritesheet = pygame.image.load(os.path.join(self.imageDir, 'boss01_core.png')).convert()
        self._glow = []
        self._glowtime = 0
        self.load_sprites()

    @property
    def anchor(self):
        return self._anchor

    @anchor.setter
    def anchor(self, a_):
        self._anchor = a_

    @property
    def anchorpos(self):
        return self._anchorpos

    @anchorpos.setter
    def anchorpos(self, ap_):
        self._anchorpos = ap_

    def load_single(self, vec4):
        rect = pygame.Rect(vec4[0], vec4[1], vec4[2], vec4[3])
        img = pygame.Surface(rect.size).convert()
        img.blit(self.spritesheet, (0, 0), rect)
        img.set_colorkey(img.get_at((0, 0)), pygame.RLEACCEL)
        return img

    def load_sprites(self):
        for image in range(4):
            x = 112 * (image % 2)
            y = 112 * math.floor(image/2)
            img = self.load_single((x,y,112,112))
            self._glow.append(img)

    def update(self):
        super().update()
        self.pos = self._anchor.pos + self._anchorpos
        self.vel = 0
        self._dmgflash = max(0, self._dmgflash - self.g.delta_time)

    def collision_bullet(self):
        col = pygame.sprite.spritecollide(self, self.l.bulletsP, False)
        if len(col) > 0:
            for bul in col:
                self.hp -= bul.dmg
                self.l.uncreate(bul.instid)
                self._dmgflash = 0.35

    def render(self):
        self._glowtime += 6.67 * self.g.delta_time
        if self._glowtime >= 7: self._glowtime = 0
        self.image = self._glow[3 - abs(3 - math.floor(self._glowtime))]
        self.l.s.blit(self.image, self.rect.copy().move(-self.l.view))



from bluebeam.objects import WhaleTurret
import os
import pygame
import math
import random


# Commander has 3 attacks
## Split Bomb - Throws bombs that split into two, the last bombs splitting twice
## Direct Burst - Fires a few bullets directly at the player in a few bursts
## Laser Wave - Creates laser waves that travel along the ground of varying max heights
# Attacks are determined randomly via a weighted RNG call
## Later, I'll implement a system where the commander tends to use attacks that have hit the player before
# Commander is also invincible - only stops attacking when the boss is destroyed

class WhaleCommander(WhaleTurret):

    def __init__(self, g_, l_, pos_=None):
        self._hanging = []
        self._gun = []
        super().__init__(g_, l_, pos_)
        self._hanging.clear()
        self._gun.clear()

        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images")
        self.spritesheet = pygame.image.load(os.path.join(self.imageDir, 'boss01_commander.png')).convert()

        #self.spritesheet = pygame.image.load(f'images/boss01_commander.png').convert()
        self.load_sprites()
        self.hp = -1
        self.rect = pygame.rect.Rect(self.pos.x, self.pos.y, 0, 0)
        self._weights = [1, 1.1, 0]
        self.nextattack = 0
        self.trackplr = 1
        self.plrpos = self.l.player.pos
        self.aimpos = pygame.math.Vector2()
        self.burst = 0
        self._shotcd = 3.75
        self.determinenext()
        self.image = self._hanging[0]

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, weights_):
        self._weights = weights_

    def load_sprites(self):
        self._hanging.append(pygame.transform.scale(self.load_single((48, 96, 48, 48)), (144, 144)))

        #tempsheet = pygame.image.load(f'images/commander_gun.png').convert()
        tempsheet = pygame.image.load(os.path.join(self.imageDir, 'commander_gun.png')).convert()
        rect = pygame.Rect(0, 0, 40, 30)
        img = pygame.Surface(rect.size).convert()
        img.blit(tempsheet, (0, 0), rect)
        img.set_colorkey(img.get_at((0, 0)), pygame.RLEACCEL)
        self._gun.append(pygame.transform.scale(img, (120,90)))


    def update(self):
        self.pos = self._anchor.pos + self._anchorpos
        self.vel = 0
        if self.trackplr == 1 and self.nextattack == 1 and self._shotcd < 1.2:
            self.plrpos = self.l.player.pos
            dist = max(abs(self.plrpos.length() - self.aimpos.length()), 0.01)
            self.aimpos = self.aimpos.lerp(self.plrpos, min(self.g.delta_time * (1000 / dist), 1))
        elif self.nextattack != 1 and self._shotcd < 1.2:
            dist = max(abs(self.plrpos.length() - self.aimpos.length()), 0.01)
            self.aimpos = self.aimpos.lerp(self.plrpos, min(self.g.delta_time * (1000 / dist), 1))
        else:
            dist = abs(pygame.math.Vector2(self.pos.x - 800, self.pos.y).length() - self.pos.length())
            self.aimpos = self.aimpos.lerp(pygame.math.Vector2(self.pos.x - 800, self.pos.y),
                                           min(self.g.delta_time * (1000 / dist), 1))

        self._shotcd = max(0, self._shotcd - self.g.delta_time)
        if self._shotcd <= 0:
            self.fire()

    def fire(self):
        if self.nextattack == 0:
            self._shotcd = 0.8
            inst = self.l.create("SplitBomb", self.pos.xy)
            inst.vel = (self.aimpos - self.pos.xy).normalize()
            inst.vel.x = inst.vel.x * 500
            inst.vel.y = inst.vel.y * 750
            inst.splits = math.floor(1 + self.burst / 2)
            angle = math.radians(random.random() * 60 + 210)
            self.plrpos = pygame.math.Vector2(self.pos.x + 800 * math.cos(angle),
                                              self.pos.y - 400 + 500 * math.sin(angle))
            self.burst += 1
            if self.burst >= 3:
                self.burst = 0
                self.determinenext()
                self._shotcd = 3.75
        elif self.nextattack == 1:
            trackplr = 0
            angle = pygame.math.Vector2().angle_to(self.aimpos - self.pos)
            inst = self.l.create("EnemyBullet", self.pos.xy)
            inst.vel = pygame.math.Vector2(400 * math.cos(math.radians(angle)), 400 * math.sin(math.radians(angle)))
            self._shotcd = 0.05
            self.burst += 1
            if self.burst % 4 == 0:
                trackplr = 1
                self._shotcd = 0.475
            if self.burst >= 12:
                trackplr = 1
                self.burst = 0
                self.determinenext()
                self._shotcd = 3.75
        else:
            self._shotcd = 0.95
            self.burst += 1
            if self.burst >= 3:
                self.burst = 0
                self.determinenext()
                self._shotcd = 3.75

    def determinenext(self):
        r = random.random() * (self.weights[0] + self.weights[1] + self.weights[2])
        if r < self.weights[0]:
            self.trackplr = 0
            self.nextattack = 0
        elif r < self.weights[0] + self.weights[1]:
            self.trackplr = 1
            self.nextattack = 1
        else:
            self.trackplr = 0
            self.nextattack = 2
        if self.nextattack == 0:
            angle = math.radians(random.random() * 30 + 240)
            self.plrpos = pygame.math.Vector2(self.pos.x + 800 * math.cos(angle),
                                              self.pos.y - 400 + 500 * math.sin(angle))

    def rotinplace(self, image, angle):
        core = image.get_rect().center
        ret = pygame.transform.rotate(image, angle)
        rec = ret.get_rect(center = ret.get_rect(center = (core[0],core[1])).center)
        return ret, rec

    def render(self):
        pygame.draw.line(self.l.s, (255, 0, 0, 100), self.pos - self.l.view, self.aimpos - self.l.view)
        self.l.s.blit(self.image, pygame.rect.Rect(self.pos.x - self.l.view.x - 72, self.pos.y - self.l.view.y - 72, 144,144))
        ang = (self.aimpos - self.pos).as_polar()[1]
        rgun = self.rotinplace(self._gun[0], 180 - ang)
        self.l.s.blit(rgun[0], rgun[1].copy().move(self.pos - self.l.view + pygame.math.Vector2(-27, -72 + 7)))


from bluebeam.objects import Enemy
import pygame
import os


class SliderEnemy(Enemy):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self.rect = pygame.rect.Rect(0, 0, 100, 100)
        self.rect.center = self.pos
        self._mhp = 50
        self.hp = self._mhp
        self.idied = 0

        self.speed = pygame.math.Vector2(900, 300)
        self.accel = 0.02
        self.gravity = 0
        self.collide_terrain = False
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images/Enemy")

        #for image/animation
        self.current_position = 0
        self.animation_counter = 0
        self.images_right = []
        self.images_left = []
        for image in range(1,5):
            img_right = pygame.image.load(os.path.join(self.imageDir,f'fish-big{image}.png')).convert_alpha()
            img_right = pygame.transform.scale(img_right, (120, 90))
            img_left = pygame.transform.flip(img_right, True, False)
            self.images_right.append(img_right)
            self.images_left.append(img_left)

        self.image = self.images_right[self.current_position]


    def update_ai_inputs(self):
        self.xmove()
        self.ymove()

    def xmove(self):
        # Should I go left or right?
        left = False
        if (self.l.player.x < self.x - 80):
            targ_xvel = -self.speed.x
            if (self.vel.x > targ_xvel):
                self.vel.x -= self.speed.x * self.accel

        elif (self.l.player.x > self.x + 80):
            targ_xvel = self.speed.x
            if (self.vel.x < targ_xvel):
                self.vel.x += self.speed.x * self.accel
                #left = False

        if (self.l.player.x < self.x):
           left = True
        else:
            left = False
        self.update_animation(left)

    def ymove(self):
        # Should I go up or down?
        if (self.l.player.y < self.y - 80):
            targ_yvel = -self.speed.y
            if (self.vel.y > targ_yvel):
                self.vel.y -= self.speed.y * self.accel

        if (self.l.player.y > self.y + 80):
            targ_yvel = self.speed.y
            if (self.vel.y < targ_yvel):
                self.vel.y += self.speed.y * self.accel

    def update_animation(self, left):
        #animation_counter = 0
        #if left:
        self.animation_counter += 1
        if self.animation_counter >= 5:
            self.animation_counter = 0
            self.current_position += 1
            if self.current_position >= 4:
                self.current_position = 0

        if left:
            self.image = self.images_left[self.current_position]
            #print(self.current_position)
        else:
            self.image = self.images_right[self.current_position]


    def update(self):
        super().update()
        if(self._active):
            self.update_ai_inputs()

    def render(self):
        # pygame.draw.rect(self.l.s,
        #                  (128 + 127 * (1 - self.idied), 128, 128 + (127 * self.idied), 100),
        #                  self.rect.copy().move(-self.l.view))

        self.l.s.blit(self.image, self.rect.copy().move(-self.l.view))
        pygame.draw.rect(self.l.s,
                         (255, 0, 0, 100),
                         pygame.rect.Rect(self.rect.topleft[0], self.rect.topleft[1] - 24, 100, 16).move(-self.l.view))
        pygame.draw.rect(self.l.s,
                         (0, 255, 0, 100),
                         pygame.rect.Rect(self.rect.topleft[0],
                                          self.rect.topleft[1] - 24,
                                          100 * (self.hp / self._mhp),
                                          16).move(-self.l.view))



# Hides "Hello from the pygame community"
from os import environ
import os

environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'

import pygame
import pytmx
from bluebeam.objects import Block
from bluebeam.objects import PhysObject, Player, YellowPower
from bluebeam.GlobalVars import GlobalVars
from bluebeam.Level import Level
from pytmx.util_pygame import load_pygame
from pygame import mixer
import math
import time

class Game_Engine:
    def __init__(self, width, height):
        # getting directory for images
        self.sourceFileDir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        self.saveDir   = os.path.join(self.sourceFileDir, "bluebeam")
        self.imageDir  = os.path.join(self.sourceFileDir, "bluebeam/images")
        self.soundDir  = os.path.join(self.sourceFileDir, "bluebeam/Sounds")
        self.buttonDir = os.path.join(self.imageDir, "Buttons")
        self.screenDir = os.path.join(self.imageDir, "Screens")
        self.savefile = open(os.path.join(self.saveDir, "savefile.txt"), 'r')

        self.game_init(width, height)

    def game_init(self, width, height):
        # System init
        self.SystemON = True
        pygame.init()
        pygame.font.init()

        self._set_resolution(width, height)

        self.clock = pygame.time.Clock()
        self.globals = GlobalVars()
        self.game_running = False

        # Screen init
        screen_size = ((self.screen_width, self.screen_height))
        self.globals.screen_size = screen_size
        self.screen = pygame.display.set_mode(screen_size, pygame.HWSURFACE|pygame.DOUBLEBUF|pygame.RESIZABLE)
        #self.screen = pygame.Surface((1920, 1080))
        pygame.display.set_caption("Blue Beam")

        
        

        # Load HUD information
        self.xscale = self.screen_width / 1920
        self.yscale = self.screen_height / 1080
        #self.xscale = 1
        #self.yscale = 1
        HUDfont = pygame.font.SysFont('Arial', 30)
        self.lives_text = HUDfont.render("Life: ", False, (0, 0, 0))
        self.heart = pygame.transform.scale(pygame.image.load(os.path.join(self.imageDir, "heart.png")).convert_alpha(), (int(50 * self.xscale), int(50 * self.yscale)))
        self.testHUD = pygame.transform.scale(pygame.image.load(os.path.join(self.screenDir, "testHUD2.png")).convert_alpha(), (self.screen_width, self.screen_height))
        self.HUDrect = self.testHUD.get_rect(center = (self.screen_width / 2, self.screen_height / 2))
        self.options = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "pause.png")).convert_alpha(), (int(370 * self.xscale), int(85 * self.yscale)))
        self.optionsRect = self.options.get_rect(center = (int(1710 * self.xscale), int(1027 * self.yscale)))
        self.menuOn = True
        self.gameOn = True

        self.level = None
        self.levelNum = None
        self.image = None

        self.pause = False
        self.dead = False
        self.restart = False
        #Load mixer/sounds
        pygame.mixer.pre_init(44100, -16, 2, 512)
        mixer.init()

        self.images = []
        self.index = 0
        self.animation_counter = 0

        self.x_scroll1 = self.screen_width / 2 - int(85 * self.xscale)
        self.x_scroll2 = self.screen_width / 2 - int(85 * self.xscale)
        self.globals.set_music_vol(5)
        self.globals.set_FX_vol(5)

        self.game_loop()
    
    def reset(self):
        # Get Level
        self.level = Level(self.globals, self.screen, self.camera_offset)

    # Get level
    def load_level(self, levelNum=None):
        if(levelNum):
            self.levelNum = levelNum
        elif( not(self.levelNum) ):
            print("Force loading level 1")
            self.levelNum = 1

        #if(self.level): self.level.delete()

        levelname = None

        if(self.levelNum == 1):
            levelname = "TestMap.tmx"
        elif(self.levelNum == 2):
            levelname = "MapLevel2.tmx"
        elif(self.levelNum == 3):
            levelname = "MapLevel3.tmx"
        self.level = Level(self.globals, self.screen, self.camera_offset, levelname)
            
    def yel_update(self):
        self.images = []
        for frame in range(1,6):
            img = pygame.image.load(os.path.join(self.imageDir, f'Yellow/frame{frame}.png'))

            #img = pygame.image.load(f'images/Yellow/frame{frame}.png')
            img = pygame.transform.scale(img, (int(80 * self.xscale), int(80 * self.yscale)))
            self.images.append(img)
        self.image = self.images[self.index]
        
        self.animation_counter += 1
        if self.animation_counter >= 10:
            self.image = self.images[self.index]
            self.index += 1
            self.animation_counter = 0
            if self.index >= 5:
                self.index = 0

    def blue_update(self):
        self.images = []
        for frame in range(1,6):
            img = pygame.image.load(os.path.join(self.imageDir, f'Blue/frame {frame}.png'))

            #img = pygame.image.load(f'images/Blue/frame {frame}.png')
            img = pygame.transform.scale(img, (int(80 * self.xscale), int(80 * self.yscale)))
            self.images.append(img)
        self.image = self.images[self.index]
        
        self.animation_counter += 1
        if self.animation_counter >= 10:
            self.image = self.images[self.index]
            self.index += 1
            self.animation_counter = 0
            if self.index >= 5:
                self.index = 0

    def red_update(self):
        self.images = []
        for frame in range(1,6):
            img = pygame.image.load(os.path.join(self.imageDir, f'Red/frame {frame}.png'))


            #img = pygame.image.load(f'images/Red/frame {frame}.png')
            img = pygame.transform.scale(img, (int(80 * self.xscale), int(80 * self.yscale)))
            self.images.append(img)
        self.image = self.images[self.index]
        
        self.animation_counter += 1
        if self.animation_counter >= 10:
            self.image = self.images[self.index]
            self.index += 1
            self.animation_counter = 0
            if self.index >= 5:
                self.index = 0

    
    def powerup_render(self):
        self.screen.blit(self.image, (int(920 * self.xscale), int(975 * self.yscale)))


    # CURRENTLY HARDCODED!!! 
    # We can modify this later - to read from another file
    def _get_settings(self):
        #self.screen_width = 1920
        #self.screen_height = 1080
        #xscale = self.screen_width / 1920
        #yscale = self.screen_height / 1080
        #self.fps = 60
        #self.camera_offset = pygame.math.Vector2(self.screen_width / 2 - int(25 * xscale), self.screen_height / 2 - int(25 * yscale))
        pass
        
    def _set_resolution(self, width, height):
        self.screen_width = width
        self.screen_height = height
        screen_size = ((self.screen_width, self.screen_height))
        self.screen = pygame.display.set_mode(screen_size)
        
        xscale = self.screen_width / 1920
        yscale = self.screen_height / 1080
        #xscale = 1
        #yscale = 1
        self.xscale = self.screen_width / 1920
        self.yscale = self.screen_height / 1080
        #self.xscale = 1
        #self.yscale = 1
        self.heart = pygame.transform.scale(pygame.image.load(os.path.join(self.imageDir, "heart.png")).convert_alpha(), (int(50 * self.xscale), int(50 * self.yscale)))
        self.testHUD = pygame.transform.scale(pygame.image.load(os.path.join(self.screenDir, "testHUD2.png")).convert_alpha(), (self.screen_width, self.screen_height))
        self.HUDrect = self.testHUD.get_rect(center = (self.screen_width / 2, self.screen_height / 2))
        self.options = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "pause.png")).convert_alpha(), (int(370 * self.xscale), int(85 * self.yscale)))
        self.optionsRect = self.options.get_rect(center = (int(1710 * self.xscale), int(1027 * self.yscale)))
        self.fps = 60
        self.camera_offset = pygame.math.Vector2(self.screen_width / 2 - int(25 * xscale), self.screen_height / 2 - int(25 * yscale))

    def _check_events(self):
        mouse = pygame.mouse.get_pos()
        if self.game_running:
            if self.level.player.hp <= 0:
                self.dead = True
                self.death_screen()
            if self.level.bossDeathState == 1:
                self.victory_screen()
                self.game_running = False
                self.menuOn = True
        
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.exit()
            elif event.type == pygame.KEYDOWN:
                if self.game_running and (event.key == pygame.K_p or event.key == pygame.K_ESCAPE):
                    self.pause = True
                    self.paused()
            elif event.type == pygame.MOUSEBUTTONUP:
                    if mouse[0] in range(self.optionsRect.left, self.optionsRect.right) and mouse[1] in range(self.optionsRect.top, self.optionsRect.bottom):
                        self.pause = True
                        self.paused()

    def _object_updates(self):
        self.level.update()
        #'''
        if self.level.player.current_power == "yellow":
            self.yel_update()
        elif self.level.player.current_power == "blue":
            self.blue_update()
        elif self.level.player.current_power == "red":
            self.red_update()
        #'''

    def _show_hud(self):
        lives = int(self.level.player.hp / 10)

        self.screen.blit(self.testHUD, self.HUDrect)
        self.screen.blit(self.options, (int(1525 * self.xscale), int(985 * self.yscale)))

        for i in range(lives):
            self.screen.blit(self.heart, (int((285 + i * 55) * self.xscale), int(1010 * self.yscale)))


        # Ranges from 0 to 2, and can go higher
        # At 1 and higher, we want an icon to show up representing "fire speed"
        # At 2 and higher, the player unlocks the charged shot mechanic - with the corresponding charge bar
        powerup_level  = self.level.player.get_powerup_level()
        
        # Ranges from 0.0 to 1.0, representing 0% to 100% charge
        charge_percent = self.level.player.get_charge_level()
        
        if self.level.player.charge_unlocked:
            pygame.draw.rect(self.screen, (255, 0, 0, 100), pygame.rect.Rect(int(1120 * self.xscale), int(1020 * self.yscale), int(350 * self.xscale), int(20 * self.yscale)))
            pygame.draw.rect(self.screen, (0, 255, 0, 100), pygame.rect.Rect(int(1120 * self.xscale), int(1020 * self.yscale), int(350 * self.xscale) * charge_percent, int(20 * self.yscale)))
        

    def _update_screen(self):
        self.level.render()
        self._show_hud()
        if self.image:
            self.powerup_render()
        # Does same thing as .flip(), but is more intuitive
        pygame.display.update()

    def text_objects(self, text, font):
        textSurface = font.render(text, True, (255, 255, 255))
        return textSurface, textSurface.get_rect()

    def _update_dt(self):
        if self.clock.get_fps() == 0:
            dt = 0
        else:
            dt = 1. / self.clock.get_fps()
        self.globals.delta_time = dt

    def _cam_pos(self):
        return self.level.player.pos - self.camera_offset

    def main_menu(self):
        self.menu_running = True
        #mixer.music.load("sounds/Dangerous Dungeon.wav")
        mixer.music.load(os.path.join(self.soundDir, "Dangerous Dungeon.wav"))
        mixer.music.set_volume(0.3 * (self.globals.get_music_vol() / 100))
        mixer.music.play(-1)

        largeText = pygame.font.SysFont("arial", int(115 * self.yscale))
        btnText = pygame.font.SysFont("arial", int(40 * self.yscale))
        TextSurf, TextRect = self.text_objects("BlueBeam", largeText)
        TextRect.center = ((self.screen_width / 2), (self.screen_height / 2) - int(200 * self.yscale))

        # Load Un-highlighted buttons
        startBtn = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "StartButton.png")).convert_alpha(), (int(300 * self.xscale), int(125 * self.yscale)))
        optBtn  = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "OptionsButton.png")).convert_alpha(), (int(300 * self.xscale), int(125 * self.yscale)))
        helpBtn  = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "HelpButton.png")).convert_alpha(), (int(300 * self.xscale), int(125 * self.yscale)))
        quitBtn = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "QuitButton.png")).convert_alpha(), (int(300 * self.xscale), int(125 * self.yscale)))

        # Load highlighted buttons
        startBtnRed = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "StartButtonRed.png")).convert_alpha(), (int(300 * self.xscale), int(125 * self.yscale)))
        optBtnRed  = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "OptionsButtonRed.png")).convert_alpha(), (int(300 * self.xscale), int(125 * self.yscale)))
        helpBtnRed  = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "HelpButtonRed.png")).convert_alpha(), (int(300 * self.xscale), int(125 * self.yscale)))
        quitBtnRed = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "QuitButtonRed.png")).convert_alpha(), (int(300 * self.xscale), int(125 * self.yscale)))

        # Load button locations        
        startRect = startBtn.get_rect(center = (self.screen_width / 2, self.screen_height / 2))
        optRect = optBtn.get_rect(center = (self.screen_width / 2, self.screen_height / 2 + int(125 * self.yscale)))
        helpRect = helpBtn.get_rect(center = (self.screen_width / 2, self.screen_height / 2 + int(250 * self.yscale)))
        quitRect = quitBtn.get_rect(center = (self.screen_width / 2, self.screen_height / 2 + int(375 * self.yscale)))
        
        while self.menu_running and self.SystemON:
            self._check_events()

            mouse = pygame.mouse.get_pos()
            click = pygame.mouse.get_pressed()

            self.screen.fill((0, 0, 0))
            self.screen.blit(TextSurf, TextRect)
            
            # Blit all buttons
            self.screen.blit(startBtn, startRect)
            self.screen.blit(optBtn, optRect)
            self.screen.blit(quitBtn, quitRect)
            self.screen.blit(helpBtn, helpRect)

            # Check if any button is hovered over / pressed
            if startRect.collidepoint(mouse):
                self.screen.blit(startBtnRed, startRect)
                if click[0] == 1:
                    self.level_select()
                    return
            if optRect.collidepoint(mouse):
                self.screen.blit(optBtnRed,  optRect)
                if click[0] == 1:
                    self.screen.fill((0, 0, 0))
                    self.options_screen()
            if helpRect.collidepoint(mouse):
                self.screen.blit(helpBtnRed, helpRect)
                if click[0] == 1:
                    self.help_screen()
            if quitRect.collidepoint(mouse):
                self.screen.blit(quitBtnRed, quitRect)
                if click[0] == 1:
                    self.exit()
                    break
            
            for event in pygame.event.get():
                if event.type == pygame.VIDEORESIZE:
                    self.screen = pygame.display.set_mode(event.size, pygame.HWSURFACE|pygame.DOUBLEBUF|pygame.RESIZABLE)

            #self.display_screen.blit(pygame.transform.scale(self.screen, (1080, 720)), (0, 0))
            pygame.display.update()
            self.clock.tick(self.fps)

    def level_select(self):
        time.sleep(0.5)
        lvlSel = True
        
        self.savefile = open(os.path.join(self.saveDir, "savefile.txt"), 'r')
        self.savefile.seek(0)
        savelines = self.savefile.readlines()
        lvl2unlocked = savelines[0].strip()
        lvl3unlocked = savelines[1].strip()

        level_scrn = pygame.transform.scale(pygame.image.load(os.path.join(self.screenDir, "LevelScreen.png")),  (self.screen_width, self.screen_height))

        backBtn = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "BackButton.png")), (int(300 * self.xscale), int(125 * self.yscale)))
        backBtnRed = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "BackButtonRed.png")), (int(300 * self.xscale), int(125 * self.yscale)))
        backRect = backBtn.get_rect(center = (int(175 * self.xscale), int(100 * self.yscale)))
        
        lvl1 = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "Level1Select.png")), (int(1671 * self.xscale), int(237 * self.yscale)))
        lvl2 = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "Level2Select.png")), (int(1671 * self.xscale), int(237 * self.yscale)))
        lvl3 = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "Level3Select.png")), (int(1671 * self.xscale), int(237 * self.yscale)))
        lvl1Red = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "Level1SelectRed.png")), (int(1671 * self.xscale), int(237 * self.yscale)))
        lvl2Red = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "Level2SelectRed.png")), (int(1671 * self.xscale), int(237 * self.yscale)))
        lvl3Red = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "Level3SelectRed.png")), (int(1671 * self.xscale), int(237 * self.yscale)))
        lvl2locked = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "Level2SelectLocked.png")), (int(1671 * self.xscale), int(237 * self.yscale)))
        lvl3locked = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "Level3SelectLocked.png")), (int(1671 * self.xscale), int(237 * self.yscale)))

        lvl1Rect = lvl1.get_rect(center = (self.screen_width / 2, int(307 * self.yscale)))
        lvl2Rect = lvl2.get_rect(center = (self.screen_width / 2, int(577 * self.yscale)))
        lvl3Rect = lvl3.get_rect(center = (self.screen_width / 2, int(847 * self.yscale)))
        
        while lvlSel == True and self.SystemON:
            self._check_events()
            mouse = pygame.mouse.get_pos()
            click = pygame.mouse.get_pressed()

            self.screen.blit(level_scrn, (0, 0))
            self.screen.blit(backBtn, backRect)
            self.screen.blit(lvl1, lvl1Rect)
            if lvl2unlocked == "1":
                self.screen.blit(lvl2, lvl2Rect)
            else:
                self.screen.blit(lvl2locked, lvl2Rect)
            if lvl3unlocked == "1":
                self.screen.blit(lvl3, lvl3Rect)
            else:
                self.screen.blit(lvl3locked, lvl3Rect)

            if backRect.collidepoint(mouse):
                self.screen.blit(backBtnRed, backRect)
                if click[0] == 1:
                    lvlSel == False
                    return
            if lvl1Rect.collidepoint(mouse):
                self.screen.blit(lvl1Red, lvl1Rect)
                if click[0] == 1:
                    mixer.music.stop()
                    mixer.music.unload()
                    self.menu_running = False
                    self.menuOn = False
                    lvlSel = False
                    self.load_level(1)
            if lvl2Rect.collidepoint(mouse) and lvl2unlocked == "1":
                self.screen.blit(lvl2Red, lvl2Rect)
                if click[0] == 1:
                    mixer.music.stop()
                    mixer.music.unload()
                    self.menu_running = False
                    self.menuOn = False
                    lvlSel = False
                    self.load_level(2)
            if lvl3Rect.collidepoint(mouse) and lvl3unlocked == "1":
                self.screen.blit(lvl3Red, lvl3Rect)
                if click[0] == 1:
                    mixer.music.stop()
                    mixer.music.unload()
                    self.menu_running = False
                    self.menuOn = False
                    lvlSel = False
                    self.load_level(3)

            pygame.display.update()
            self.clock.tick(self.fps)

    def button(self, x, y, width, height, color, action = None):
        global pen_color, pen_size
        mouse = pygame.mouse.get_pos()
        click = pygame.mouse.get_pressed()
        if x + width > mouse[0] > x and y + height + 12 > mouse[1] > y - 12:
            pygame.draw.rect(self.screen, color, (x, y, width, height))
            if click[0] == 1 and action != None:
                if action == "scroll1":
                    self.x_scroll1 = mouse[0]
                if action == "scroll2":
                    self.x_scroll2 = mouse[0]
        else:
            pygame.draw.rect(self.screen, color, (x, y, width, height))
    
    
    def options_screen(self):
        pauseFont = pygame.font.SysFont('verdana', int(55 * self.yscale))

        back, backRect = self.text_objects("Back", pauseFont)
        backRect.center = (self.screen_width / 2, int(400 * self.yscale))

        arrowLeft = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "ArrowLeft.png")).convert_alpha(), (int(75 * self.xscale), int(75 * self.yscale)))
        arrowRight = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "ArrowRight.png")).convert_alpha(), (int(75 * self.xscale), int(75 * self.yscale)))
        arrowLeftRect = arrowLeft.get_rect(center = (int(800 * self.xscale), int(683 * self.yscale)))
        arrowRightRect = arrowRight.get_rect(center = (int(1230 * self.xscale), int(683 * self.yscale)))

        options = True
        if self.screen_width == 1080:
            resOpt1 = True
            resOpt2 = False
        elif self.screen_width == 1920:
            resOpt1 = False
            resOpt2 = True

        
        while options == True and self.SystemON:
            mouse = pygame.mouse.get_pos()
            
            self.screen.blit(pygame.transform.scale(pygame.image.load(os.path.join(self.screenDir, "OptionsMenu.png")).convert_alpha(), (self.screen_width, self.screen_height)), (0, 0))
            self.screen.blit(back, backRect)

            self.button(self.screen_width / 2 - int(100 * self.xscale), self.screen_height / 2 - int(50 * self.yscale), int(275 * self.xscale), int(4 * self.yscale), (255, 255, 255), "scroll1")
            pygame.draw.rect(self.screen, (255, 0, 0), (self.x_scroll1 - int(5 * self.xscale), self.screen_height / 2 - int(65 * self.yscale), int(10 * self.xscale), int(30 * self.yscale)))
            self.globals.set_music_vol(math.trunc((self.x_scroll1 - (self.screen_width / 2 - int(100 * self.xscale))) / int(275 * self.xscale) * 100))
            if self.globals.get_music_vol() == 98 or self.globals.get_music_vol() == 99:
                self.globals.set_music_vol(100)
            mixer.music.set_volume(0.3 * (self.globals.get_music_vol() / 100))

            self.button(self.screen_width / 2 - int(100 * self.xscale), self.screen_height / 2 + int(45 * self.yscale), int(275 * self.xscale), int(4 * self.yscale), (255, 255, 255), "scroll2")
            pygame.draw.rect(self.screen, (255, 0, 0), (self.x_scroll2 - int(5 * self.xscale), self.screen_height / 2 + int(30 * self.yscale), int(10 * self.xscale), int(30 * self.yscale)))
            self.globals.set_FX_vol(math.trunc((self.x_scroll2 - (self.screen_width / 2 - int(100 * self.xscale))) / int(275 * self.xscale) * 100))
            if self.globals.get_FX_vol() == 98 or self.globals.get_FX_vol() == 99:
                self.globals.set_FX_vol(100)

            volume, volumeRect = self.text_objects("Music Volume", pauseFont)
            (volumeRect.top, volumeRect.left) = (self.screen_height / 2 - int(85 * self.yscale), int(325 * self.xscale))

            sound, soundRect = self.text_objects(str(self.globals.get_music_vol()), pauseFont)
            soundRect.center = (int(1200 * self.xscale), self.screen_height / 2 - int(50 * self.yscale))

            fx, fxRect = self.text_objects("Effects Volume", pauseFont)
            (fxRect.top, fxRect.left) = (self.screen_height / 2 + int(10 * self.xscale), int(325 * self.xscale))

            fxsound, fxsoundRect = self.text_objects(str(self.globals.get_FX_vol()), pauseFont)
            fxsoundRect.center = (int(1200 * self.xscale), self.screen_height / 2 + int(45 * self.yscale))

            res, resRect = self.text_objects("Resolution", pauseFont)
            (resRect.top, resRect.left) = (self.screen_height / 2 + int(105 * self.yscale), int(325 * self.xscale))

            self.screen.blit(volume, volumeRect)
            self.screen.blit(sound, soundRect)
            self.screen.blit(fx, fxRect)
            self.screen.blit(fxsound, fxsoundRect)
            
            """ if not self.game_running:
                self.screen.blit(res, resRect)
                self.screen.blit(arrowLeft, arrowLeftRect)
                self.screen.blit(arrowRight, arrowRightRect)

                
                if resOpt1:
                    resText, resTextRect = self.text_objects("1080x720", pauseFont)
                    (resTextRect.top, resTextRect.left) = (self.screen_height / 2 + int(105 * self.yscale), int(875 * self.xscale))
                    self.screen.blit(resText, resTextRect)
                elif resOpt2:
                    resText, resTextRect = self.text_objects("1920x1080", pauseFont)
                    (resTextRect.top, resTextRect.left) = (self.screen_height / 2 + int(105 * self.yscale), int(860 * self.xscale))
                    self.screen.blit(resText, resTextRect) """

            if backRect.collidepoint(mouse):
                self.screen.blit(pauseFont.render("Back", True, (255, 0, 0)), backRect)
            
            for event in pygame.event.get():
                            
                if event.type == pygame.MOUSEBUTTONUP:
                    if backRect.collidepoint(mouse):
                        if not self.game_running:    
                            if resOpt1 and not self.screen_width == 1080:
                                pygame.quit()
                                self.game_init(1080, 720)
                            elif resOpt2 and not self.screen_width == 1920:
                                pygame.quit()
                                self.game_init(1920, 1080)
                        options = False
                    if not self.game_running:    
                        if arrowLeftRect.collidepoint(mouse) or arrowRightRect.collidepoint(mouse):
                            resOpt1 = not resOpt1
                            resOpt2 = not resOpt2

            pygame.display.update()
            self.clock.tick(60)
    
    def help_screen(self):
        pauseFont = pygame.font.SysFont('verdana', int(55 * self.yscale))

        resume, resumeRect = self.text_objects("Resume", pauseFont)
        resumeRect.center = (self.screen_width / 2, int(400 * self.yscale))

        help_scrn1 = pygame.transform.scale(pygame.image.load(os.path.join(self.screenDir, "HelpScreen1.png")).convert_alpha(), (self.screen_width, self.screen_height))
        help_scrn2 = pygame.transform.scale(pygame.image.load(os.path.join(self.screenDir, "HelpScreen2.png")).convert_alpha(), (self.screen_width, self.screen_height))
        help_scrn3 = pygame.transform.scale(pygame.image.load(os.path.join(self.screenDir, "HelpScreen3.png")).convert_alpha(), (self.screen_width, self.screen_height))


        exitBtn = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "ExitButton.png")), (int(300 * self.xscale), int(125 * self.yscale)))
        exitBtnRed = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "ExitButtonRed.png")), (int(300 * self.xscale), int(125 * self.yscale)))
        exitRect = exitBtn.get_rect(center = (self.screen_width / 2 + int(700 * self.xscale), self.screen_height / 2 + int(425 * self.yscale)))

        #helpBtn = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "HelpButton.png")), (300, 125))
        #helpBtnRed = pygame.transform.scale(pygame.image.load(os.path.join(self.buttonDir, "HelpButtonRed.png")), (300, 125))
        #helpRect = helpBtn.get_rect(center = (self.screen_width / 2 + 400, self.screen_height / 2 + 200))
        
        help = True
        scr1 = True
        scr2 = False
        scr3 = False

        
        while help and self.SystemON:
            mouse = pygame.mouse.get_pos()
            if scr1 == True:
                self.screen.blit(help_scrn1, (0, 0))
                self.screen.blit(exitBtn, exitRect)

                if exitRect.collidepoint(mouse):
                    self.screen.blit(exitBtnRed, exitRect)
                for event in pygame.event.get():
                    if event.type == pygame.MOUSEBUTTONUP:
                        if exitRect.collidepoint(mouse):
                            scr1 = False
                            scr2 = True
                pygame.display.update()
                self.clock.tick(60)
            elif scr2 == True:
                self.screen.blit(help_scrn2, (0, 0))
                self.screen.blit(exitBtn, exitRect)

                if exitRect.collidepoint(mouse):
                    self.screen.blit(exitBtnRed, exitRect)
                for event in pygame.event.get():      
                    if event.type == pygame.MOUSEBUTTONUP:
                        if exitRect.collidepoint(mouse):
                            scr2 = False
                            scr3 = True
                pygame.display.update()
                self.clock.tick(60)
            elif scr3 == True:
                self.screen.blit(help_scrn3, (0, 0))
                self.screen.blit(exitBtn, exitRect)

                if exitRect.collidepoint(mouse):
                    self.screen.blit(exitBtnRed, exitRect)
                for event in pygame.event.get():
                    if event.type == pygame.MOUSEBUTTONUP:
                        if exitRect.collidepoint(mouse):
                            scr3 = False
                            help = False
                            self.screen.fill((0, 0, 0))
                pygame.display.update()
                self.clock.tick(60)

    def game_loop(self):
        while self.gameOn and self.SystemON:
            if self.menuOn:
                self.main_menu()
            elif self.restart:
                self.restart = False
                self.load_level()
                self.run_game()
            else:
                self.run_game()

    def run_game(self):
        #print(self.sourceFileDir)
        self.game_running = True
        # Get Level
        while self.game_running and self.SystemON:   
            if self.pause == False and self.dead == False and self.restart == False:
                self._update_dt()
                self._check_events()
                self._object_updates()
            self._update_screen()
            self.clock.tick(self.fps) # Performs a time.sleep, at an interval ensuring the game runs at (fps)
        self.level.delete()

    
    def paused(self):
        pause_scrn = pygame.transform.scale(pygame.image.load(os.path.join(self.imageDir, "PauseMenu.png")).convert_alpha(), (self.screen_width, self.screen_height))

        pauseFont = pygame.font.SysFont('verdana', int(55 * self.yscale))

        resume, resumeRect = self.text_objects("Resume", pauseFont)
        opt, optRect = self.text_objects("Options", pauseFont)
        help, helpRect = self.text_objects("Help", pauseFont)
        menu, menuRect = self.text_objects("Main Menu", pauseFont)
        end, endRect = self.text_objects("Quit", pauseFont)

        resumeRect.center = (self.screen_width / 2, int(400 * self.yscale))
        optRect.center = (self.screen_width / 2, int(475 * self.yscale))
        helpRect.center = (self.screen_width / 2, int(550 * self.yscale))
        menuRect.center = (self.screen_width / 2, int(625 * self.yscale))
        endRect.center = (self.screen_width / 2, int(700 * self.yscale))
        

        
        while self.pause == True and self.SystemON:
            mouse = pygame.mouse.get_pos()
            click = pygame.mouse.get_pressed()

            self.screen.blit(pause_scrn, (0, 0))
            self.screen.blit(resume, resumeRect)
            self.screen.blit(opt, optRect)
            self.screen.blit(help, helpRect)
            self.screen.blit(menu, menuRect)
            self.screen.blit(end, endRect)


            if resumeRect.collidepoint(mouse):
                self.screen.blit(pauseFont.render("Resume", True, (255, 0, 0)), resumeRect)
            if optRect.collidepoint(mouse):
                self.screen.blit(pauseFont.render("Options", True, (255, 0, 0)), optRect)
            if helpRect.collidepoint(mouse):
                self.screen.blit(pauseFont.render("Help", True, (255, 0, 0)), helpRect)
            if menuRect.collidepoint(mouse):
                self.screen.blit(pauseFont.render("Main Menu", True, (255, 0, 0)), menuRect)
            if endRect.collidepoint(mouse):
                self.screen.blit(pauseFont.render("Quit", True, (255, 0, 0)), endRect)
            
            for event in pygame.event.get():
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        self.pause = False
                            
                elif event.type == pygame.MOUSEBUTTONUP:
                    if resumeRect.collidepoint(mouse):
                        self.pause = False
                    if optRect.collidepoint(mouse):
                        self.options_screen()
                    if helpRect.collidepoint(mouse):
                        self.help_screen()
                    if menuRect.collidepoint(mouse):
                        self.game_running = False
                        self.pause = False
                        self.menuOn = True
                    if endRect.collidepoint(mouse):
                        self.exit()

            pygame.display.update()
            self.clock.tick(60)
            

    def death_screen(self):
        
        restart = pygame.transform.scale(pygame.image.load(os.path.join(self.imageDir, "RestartButton.png")), (int(300 * self.xscale), int(125 * self.yscale)))
        restartRect = restart.get_rect(center = (self.screen_width / 2, self.screen_height / 2 - int(25 * self.yscale)))

        menuBtn = pygame.transform.scale(pygame.image.load(os.path.join(self.imageDir, "MenuButton.png")), (int(300 * self.xscale), int(125 * self.yscale)))
        menuRect = menuBtn.get_rect(center = (self.screen_width / 2, self.screen_height / 2 + int(100 * self.yscale)))

        quitBtn = pygame.transform.scale(pygame.image.load(os.path.join(self.imageDir, "QuitButton.png")), (int(300 * self.xscale), int(125 * self.yscale)))
        quitRect = quitBtn.get_rect(center = (self.screen_width / 2, self.screen_height / 2 + int(225 * self.yscale)))

        pygame.mixer.music.stop()
        # pygame.mixer.music.load(f'Sounds/cherish.xm')
        cherish_sound_dir = os.path.join(self.soundDir, f'cherish.xm')
        pygame.mixer.music.load(cherish_sound_dir)
        pygame.mixer.music.play(-1,0,0)

        deathscreen = pygame.image.load(os.path.join(self.imageDir, "DeathScreen.png")).convert_alpha()
        deathscreen = pygame.transform.scale(deathscreen, (self.screen_width, self.screen_height))
        
        while self.dead and self.SystemON:
            mouse = pygame.mouse.get_pos()

            self.screen.blit(deathscreen, (0, 0))
            self.screen.blit(restart, restartRect)
            self.screen.blit(menuBtn, menuRect)
            self.screen.blit(quitBtn, quitRect)

            if restartRect.collidepoint(mouse):
                self.screen.blit(pygame.transform.scale(pygame.image.load(os.path.join(self.imageDir, "RestartButtonRed.png")), (int(300 * self.xscale), int(125 * self.yscale))),
                                 restartRect)
            if menuRect.collidepoint(mouse):
                self.screen.blit(pygame.transform.scale(pygame.image.load(os.path.join(self.imageDir, "MenuButtonRed.png")), (int(300 * self.xscale), int(125 * self.yscale))),
                                 menuRect)
            if quitRect.collidepoint(mouse):
                self.screen.blit(pygame.transform.scale(pygame.image.load(os.path.join(self.imageDir, "QuitButtonRed.png")), (int(300 * self.xscale), int(125 * self.yscale))),
                                 quitRect)

            for event in pygame.event.get():
                if event.type == pygame.MOUSEBUTTONUP:
                    if mouse[0] in range(restartRect.left, restartRect.right) and mouse[1] in range(restartRect.top, restartRect.bottom):
                        self.game_running = False
                        self.dead = False
                        self.restart = True
                    if mouse[0] in range(menuRect.left, menuRect.right) and mouse[1] in range(menuRect.top, menuRect.bottom):
                        self.game_running = False
                        self.dead = False
                        self.menuOn = True
                    if mouse[0] in range(quitRect.left, quitRect.right) and mouse[1] in range(quitRect.top, quitRect.bottom):
                        self.exit()

            pygame.display.update()
            self.clock.tick(60)

    def victory_screen(self):
        if self.levelNum == 1:
            self.savefile.seek(0)
            savelines = self.savefile.readlines()
            lvl2unlocked = savelines[0].strip()
            lvl3unlocked = savelines[1].strip()
            self.savefile = open(os.path.join(self.saveDir, "savefile.txt"), 'w')
            self.savefile.writelines(["1\n", lvl3unlocked])
        if self.levelNum == 2:
            self.savefile.seek(0)
            self.savefile = open(os.path.join(self.saveDir, "savefile.txt"), 'w')
            self.savefile.writelines(["1\n", "1"])



        victFont = pygame.font.SysFont('Verdana', int(120 * self.yscale))
        start = time.time()

        while time.time() - start < 5:
            if int(time.time() - start) % 2 == 0:
                self.screen.blit(victFont.render("LEVEL CLEAR", True, (255, 0, 0)), (self.screen_width / 2 - int(400 * self.xscale), self.screen_height / 2 - int(300 * self.yscale)))
            else:
                self.screen.blit(victFont.render("LEVEL CLEAR", True, (0, 255, 0)), (self.screen_width / 2 - int(400 * self.xscale), self.screen_height / 2 - int(300 * self.yscale)))
            pygame.display.update()
            self.clock.tick(self.fps)
        self.level.bossDeathState = 0

    def exit(self):
        self.SystemON = False
        self.gameOn = False
        self.pause = False
        self.dead = False
        self.menu_running = False
        self.game_running = False
        #pygame.quit()

def main():
    engine = Game_Engine(1920, 1080)
    #engine.game_loop()


if __name__ == '__main__':
    main()

    # engine = Game_Engine()
    # engine.game_loop()
    #engine.run_game()



from bluebeam.objects import Enemy
import pygame, os
from pygame import Vector2 as vec2


class GammaCore(Enemy):

    def __init__(self, g_, l_, pos_):
        super().__init__(g_, l_, pos_)
        self.body = None
        self.exposed = 0
        self.door = None
        self.heart = []
        self.drect = pygame.rect.Rect(0,0,124,116)
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imagedir = os.path.join(os.path.dirname(self.sourceDir), "images/Boss/GammaGenerator")
        self.spritesheet = pygame.image.load(os.path.join(self.imagedir, 'boss03_core_open.png')).convert()
        self.load_sprites()
        self.hp = 1
        self.rect = pygame.rect.Rect(0,0,192,44)
        self.pos = self.rect.center = pos_
        self.depth = 9
        self.flash = 0

    def load_sprites(self):
        for i in range(0,4):
            x = i*48
            img = self.load_single((x,0,48,26))
            self.heart.append(pygame.transform.scale(img,(192,104)))
        self.spritesheet = pygame.image.load(os.path.join(self.imagedir, 'boss03_core_shut.png')).convert()
        self.door = pygame.transform.scale(self.load_single((0,0,31,29)),(124,116))

    def setup(self, body_):
        self.body = body_

    def update(self):
        self.flash = (self.flash + (7 + 5.25*(3-self.body.phase)) * self.g.delta_time) % 7
        if self.exposed > 0:
             self.exposed = max(0, self.exposed - self.g.delta_time)
        self.pos = self.rect.center = self.body.pos + vec2(0, 112)
        if self.exposed > 0:
            self.collision_bullet()
        pass

    def collision_bullet(self):
        col = pygame.sprite.spritecollide(self, self.l.bulletsP, False)
        if len(col) > 0:
            for bul in col:
                self.body.hp -= bul.dmg
                bul.hit()

    def render(self):
        item = 3-abs(int(self.flash)-3)
        if self.exposed == 0:
            self.drect.center = self.pos + vec2(-62,0)
            self.l.s.blit(self.door, self.drect.copy().move(-self.l.view))
            self.drect.center = self.pos + vec2(62,0)
            self.l.s.blit(pygame.transform.flip(self.door,True,False), self.drect.copy().move(-self.l.view))
        else:
            self.l.s.blit(self.heart[item], self.rect.copy().move(-self.l.view - vec2(0,30)))
            self.drect.center = self.pos + vec2(-(62 + min(96., 96.*(6.5-self.exposed))),0)
            self.l.s.blit(self.door, self.drect.copy().move(-self.l.view))
            self.drect.center = self.pos + vec2((62 + min(96., 96.*(6.5-self.exposed))),0)
            self.l.s.blit(pygame.transform.flip(self.door,True,False), self.drect.copy().move(-self.l.view))
        pass



class GlobalVars:

    def __init__(self):
        self._tick_rate = -1
        self._delta_time = -1
        self._instid = 100000
        self._boss1_death = 0

    @property
    def screen_size(self):
        return self._screen_size

    @screen_size.setter
    def screen_size(self, sz):
        self._screen_size = sz

    @property
    def tick_rate(self):
        return self._tick_rate

    @tick_rate.setter
    def tick_rate(self, tr):
        self._tick_rate = tr

    @property
    def delta_time(self):
        return self._delta_time

    @delta_time.setter
    def delta_time(self, dt):
        self._delta_time = dt

    @property
    def instid(self):
        return self._instid

    @instid.setter
    def instid(self, id):
        self._instid = id

    def getid(self):
        self._instid += 1
        return self._instid - 1

    def set_FX_vol(self, volume):
        self.FXVol = volume

    def set_music_vol(self, volume):
        self.musicVol = volume

    def get_FX_vol(self):
        return self.FXVol
    
    def get_music_vol(self):
        return self.musicVol



import math
from bluebeam.objects import Enemy
import pygame
from pygame import Vector2 as vec2
import os
import random

class SpireBody(Enemy):
    phase = 4  # phases count in reverse, starts at 4 for the first frame
    body = pygame.sprite.Sprite()

    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self._bosscore = None
        self._sprites = []
        self._glass = []
        self._streams = []
        self._commander = []
        self._extensions = []
        self._jets = []
        self.depth = 16
        self.rect = pygame.rect.Rect(self.pos.x, self.pos.y, 162, 480)
        self.rect.center = self.pos
        self.image = None
        self.contactdmg = 10
        self.pushforce = 960
        self._anims = 0
        self.plrtookdmg = 0
        self.plrlasthp = self.l.player.hp

        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imagedir = os.path.join(os.path.dirname(self.sourceDir), "images/Boss/BetaSpire")
        self.spritesheet = pygame.image.load(os.path.join(self.imagedir, 'boss02_body.png')).convert()
        self.load_sprites()

        self.activeextents = 1
        self.extentoffset = 0
        self.phasedownexplotime = 0.1
        self.phasedownexplo = 0
        pass

    def load_sprites(self):
        # Body
        for image in range(3):
            x = 72 * (image % 3)
            y = 0
            img = self.load_single((x,y,72,160))
            self._sprites.append(pygame.transform.scale(img, (216,480)))
        self.image = self._sprites[0]
        # Streams
        self.spritesheet = pygame.image.load(os.path.join(self.imagedir, 'boss02_streams.png'))
        for image in range(4):
            x = 48 * (image % 4)
            y = 0
            img = self.load_single((x,y,48,124))
            self._streams.append(pygame.transform.scale(img, (144,372)))
        # Commander Pilot
        self.spritesheet = pygame.image.load(os.path.join(self.imagedir, 'boss02_commander.png'))
        for image in range(4):
            x = 36 * (image % 2)
            y = 20 * int(image/2)
            img = self.load_single((x,y,36,20))
            self._commander.append(pygame.transform.scale(img, (108,60)))
        # Glass
        self.spritesheet = pygame.image.load(os.path.join(self.imagedir, 'boss02_glass.png'))
        for image in range(3):
            x = 0
            y = 27 * image
            img = self.load_single((x,y,48,27))
            self._glass.append(pygame.transform.scale(img, (144,81)))
        # Extensions
        self.spritesheet = pygame.image.load(os.path.join(self.imagedir, 'boss02_extents.png'))
        # Turret
        img = self.load_single((0, 0, 16, 16))
        self._extensions.append(pygame.transform.scale(img, (48, 48)))
        # Rocket Battery
        img = self.load_single((16, 0, 24, 16))
        self._extensions.append(pygame.transform.scale(img, (72, 48)))
        # BulletBarrage Anti-Cheese Laser
        img = self.load_single((0, 16, 24, 24))
        self._extensions.append(pygame.transform.scale(img, (72, 72)))
        # Jets
        img = self.load_single((24, 16, 16, 24))
        self._extensions.append(pygame.transform.scale(img, (48, 72)))
        # Laser Cannon
        img = self.load_single((0, 40, 40, 24))
        self._extensions.append(pygame.transform.scale(img, (120, 72)))
        # Flames
        self.spritesheet = pygame.image.load(os.path.join(self.imagedir, 'boss02_jets.png'))
        for i in range(0,3):
            img = self.load_single((12*i, 0, 12, 16))
            self._jets.append(pygame.transform.scale(img, (36, 48)))
            img = self.load_single((24*i, 16, 24, 28))
            self._jets.append(pygame.transform.scale(img, (72, 84)))
            if i == 1:
                self._jets.append(self._jets[0])
                self._jets.append(self._jets[1])

    def movement(self):
        self.pos = self._bosscore.pos + vec2(0,45)
        self.rect.center = self.pos

    def update(self):
        if self._active:
            self.plrtookdmg = max(0, self.plrtookdmg - self.g.delta_time)
            if self.plrlasthp != self.l.player.hp:
                self.plrtookdmg = 0.8
            self.plrlasthp = self.l.player.hp
            if self.rect.colliderect(self.l.player.rect):
                self.hit()
            if self.phasedownexplo > 6:
                self.image = self._sprites[min(max(0,4-self._bosscore.phase), 2)]
            else:
                self.image = self._sprites[min(max(0, 3 - self._bosscore.phase), 2)]
            if self.phasedownexplo > 0:
                self.phasedownexplotime = max(0,self.phasedownexplotime - self.g.delta_time)
                if self.phasedownexplotime <= 0:
                    for i in range(0,3):
                        boom = self.l.create("Explosion", self.pos + pygame.math.Vector2(random.randint(-96,96),random.randint(-128,256)))
                        boom.dmg = 0
                        boom.radius(96)
                        boom.depth = 64
                    self.phasedownexplo -= 1
                    self.phasedownexplotime = 0.1
        else:
            self.check_active()

    def hit(self):
        if (self.contactdmg > 0):
            self.l.player.take_damage(self.contactdmg)
        furthest = 0
        if (abs(self.l.player.pos.x - self._bosscore._popuppts[0].x) < abs(self._bosscore._popuppts[1].x - self.l.player.pos.x)):
            furthest = 1
        else: furthest = -1
        self.l.player.pos += pygame.math.Vector2(self.pushforce * self.g.delta_time * furthest, -self.pushforce/3 * self.g.delta_time)


    def render(self):
        if self._active:
            self._anims = (self._anims + self.g.delta_time / 0.0625) % 8
            # Draw commander
            x = 27
            y = 33
            mode = 0
            if self.phasedownexplo > 0:
                mode = 3
            elif self.plrtookdmg > 0:
                mode = 1 + int(self._anims/2) % 2
            self.l.s.blit(self._commander[mode], self.rect.copy().move(-self.l.view + pygame.math.Vector2(x, y)))
            # Draw glass
            x = 9
            y = 9
            self.l.s.blit(self._glass[min(3-self._bosscore.phase, 2)], self.rect.copy().move(-self.l.view + pygame.math.Vector2(x, y)))
            if self.extentoffset != 0:
                if self.activeextents == 0 or self.activeextents == 4:

                    # Draw turrets
                    x = self.rect.width/2 + (self.rect.width/2 - 18 - 48*(1 - self.extentoffset)) * self._bosscore.playerdir + (-48*(self._bosscore.playerdir==-1))
                    for i in range(0,6):
                        y = self._bosscore._turretpoints[i].y + 212
                        img = self._extensions[0]
                        if self._bosscore.playerdir == 1:
                            img = pygame.transform.flip(img,True,False)
                        self.l.s.blit(img, self.rect.copy().move(-self.l.view + pygame.math.Vector2(x,y)))

                    # Draw laser emitter
                    x = self.rect.width/2 + (self.rect.width/2 - 36 - 24*(1-self.extentoffset)) * self._bosscore.playerdir
                    y = 36 + 72*(1-self.extentoffset)
                    img = self._extensions[2]
                    if self._bosscore.playerdir == -1:
                        img = pygame.transform.flip(img, True, False)
                    self.l.s.blit(img, self.rect.copy().move(-self.l.view + pygame.math.Vector2(x,y)))
                elif self.activeextents == 1:
                    # Draw rocket turrets
                    for i in range(0,2):
                        x = self._bosscore._rocketpoints[i].x
                        y = self._bosscore._rocketpoints[i].y + 212 + 72 * (1 - self.extentoffset)
                        img = self._extensions[1]
                        if i == 0:
                            img = pygame.transform.flip(img, True, False)
                        self.l.s.blit(img, self.rect.copy().move(-self.l.view + pygame.math.Vector2(x, y)))
                elif self.activeextents == 2:
                    # Draw sweep jets and flames
                    x = self.rect.width/2 - 12 + (self.rect.width/2 - 48*(1 - self.extentoffset)) * -self._bosscore.playerdir + (-24*(self._bosscore.playerdir==1))
                    for i in range(0,3):
                        y = self._bosscore._jetpoints[i].y + 212
                        offset = pygame.math.Vector2(
                            0 - 36*(self._bosscore._sweep > 0) + 33*(self._bosscore._sweep > 0 and self._bosscore.playerdir==-1),
                            12 - 18*(self._bosscore._sweep > 0)
                        )
                        img = self._extensions[3]
                        if self._bosscore.playerdir == -1:
                            img = pygame.transform.flip(img,True,False)
                        self.l.s.blit(img, self.rect.copy().move(-self.l.view + pygame.math.Vector2(x, y)))
                        if 4-i > self._bosscore._attackburst:
                            img = self._jets[2*int(math.floor(self._anims) % 4) + (self._bosscore._sweep > 0)]
                            if self._bosscore.playerdir == -1:
                                img = pygame.transform.flip(img, True, False)
                            self.l.s.blit(img, self.rect.copy().move(-self.l.view + offset + pygame.math.Vector2(x - (-9*(self._bosscore.playerdir==1) + 36*self._bosscore.playerdir), y)))
            self.l.s.blit(self.image, self.rect.copy().move(-self.l.view - pygame.math.Vector2(27,0)))
            self.l.s.blit(self._streams[int(self._anims/2) % 4], self.rect.copy().move(-self.l.view + pygame.math.Vector2(9, 99)))
